! Module: omdetect_mod
!
! Authors:  Tim Harries - MSSL (orig) 
!           almost completely rewritten by J. Mittaz MSSL
!          
! Description: Uses a simple algorithm to detect sources in an OM OSW image
! 
! Comments: 
! 
! $Id: module.f90,v 1.5 1998/01/13 08:39:15 gvacanti Exp $
!
! History
!
!    Created: Jan 1998 (TJH)
! 
!        July 1998: completely revised algorithm employed
!
!
!    27 April 1998: Removed diagnostic output
!                 : changed copy from rimage to image
!                 : removed datestamp call (TJH)
!
!    07 May   1998: Correct code for count rate bug (TJH)
!
!    Nov 2000 : Background by smoothing + Poisson noise for detection signif
!    Jan 2000 : Now uses mulitple scales + extractor type source 
!               parameterisation including deblending
!    24th August 2000 - AMK
!    Code added to read in the quality array.
!    The task will have to be modified eventually so that the quality array is processed.
!    We have seen images that have been generated by OMFLATFIELD that contain pixel values 
!    that contain values that have been set to the BLANK value. This occurrs due to the Flatfield
!    being zero for certain pixels. the quality array has been flagged to the value of 16 for these
!    cases. The image has been modified to contain the value zero as a temporary fix.
!    2nd. October 2000 - AMK
!    Reinstated the call to CAL_omGetPlateScale
!    3rd. Octeober 2000 - AMK
!    Replace the call getFIlterID in mssllib by the OAL call filterIdToInt
!    10th October 2000 - AMK Added White to list of filters - this can occur in the Summary file now
!    11th December (CB) - Background counts for each source added to the source list.
!    1st May 2001 (CB) - All pgplot drawing code removed. Limiting plate count rate calculated using 
!                        the photometry aperture radius. Code tidied up.
!    28th May 2001 (VNY) - an error was corrected in the getMoments routine (file omdetectsourceparam_mod.f90),
!                        (indexation was omitted for srcParam array in one of the assignation) 
!    06 June 2001 (CB) - Added output column to store the aperture radius for each source 
!    04 July 2001 (VNY) - Protection against subscription of range for histogram is added in computeMode routine
!
!    09 July 2001 (VNY) - In order to avoid a problem with the CAL routine (ComputePSFWidthths) when 
!                         the choisen filter is 20 (MAGNIFIER), this routine is not called for the 
!                         case nSources=0 (no detected sources within the FOW). The same is done for 
!                         CAL_omGetApertureRadius().
!    12 July 2001 (VNY) - smoothSize parameter has been used in the first call to smoothImage routine
!                         instead of a fixed value 3.0 used before. With this fixed value FAST chain 
!                         sometimes failed to detect images
!    16 July 2001 (VNY) - square root has been added in the calculation of distances between sources when 
!                         analyzing overlapping sources. The tolerance for this distance is taken max of the 
!                         semi-major axis of the source ellipses
!    26th July 2001 (CB) First call to subroutine smoothImage now uses a smoothSize parameter of 3.0 (as originall)
!
!    2nd August 2001 (VNY) - In order to improve the program's detection ability, negative values appearing
!                         in the image array after background substraction have been zeroed
!                         (omdetectsourceparam_mod.f90 file)
! 
! 
!    6th of August 2001 (VNY) A code in the COMPUTEMODE routine which  underestimates the background median value,
!                          nas been removed.
!                          Parameter TOLERANCE has been set to 5 pixel fixed value (elimination of 
!                          multiple sources) 
!    7th of August 2001 (VNY) - omMode variable has been introduced (to be used for the differences 
!                          in the algorithm between FAST and IMAGING modes of the Optical Monitor
!    29th of August 2001 (VNY) - a correction of XY-position of the source at the edge of FOV (only FAST mode)
!
!    4th of September 2001 (VNY) - modified correction of XY-position of the source for the FAST mode
!    20th of September 2001 (VNY) - a sourcePars array indexation bug has been fixed 
!    23rd of October 2001 (VNY) - a protection against negative values for backgroundVal and imageVal
!                          variables has been introduced before calculation of log and lnFactrl in order
!                          to improve robustness of the programme
!    24th of October 2001 (VNY) - additional check has been introduced for the HISTOGRAM array allocation
!    25th of October 2001 (VNY) - additional protection agains too high values of nhist has been introduced
!                          before allocation of the HISTOGRAM array
!*********************************************************************************************************** 
module omdetect_mod

    use dal
    use cal
    use param  
!    use sort_mod
    use sasdal
    use omdetectnumerics_mod
    use omdetectsourceparams_mod
    use omdetectphotometry_mod
    use get_free_io_unit_mod
    use getFilterID_mod
    use errorHandling

    implicit none 
 
  private
  public  omdetect

    type srcpixel                             ! Pixel list node.
        integer(kind=int32) x                  ! X position.
        integer(kind=int32) y                  ! Y position.
        type(srcpixel), pointer :: next        ! Next pixel in source.
        type(srcpixel), pointer :: previous    ! Previous one (mem removal)
    end type srcpixel
  
    type src                                   ! Source list node.
        integer(kind=int32) sourceNum          ! Source number.
        integer(kind=int32) numPixels          ! Number of pixels.
        type(srcpixel), pointer :: pixelList   ! Pixel list.
        type(src), pointer :: next             ! Next source in list.
        type(src), pointer :: previous         ! previous one (memory removal)
    end type src


!
!---------------------------------------------------------------------
!
  
contains 
  
!---------------------------------------------------------------------
subroutine omdetect
!---------------------------------------------------------------------
    !*******************
    ! Filenames
    !*******************
    character(len=1024)  :: set                     ! filename of input image
    character(len=1024)  :: backgroundImage         ! Output background image
    character(len=1024)  :: levelImageName
    character(len=1024)  :: outset                  ! filename of osw source list
    character(len=1024)  :: flatField               ! filename of osw flat field
    character(len=1024)  :: mod8noise               ! filename of mod8noise
    character(len=1024)  :: regionfile              ! filename of mod8noise

    !****************
    ! DAL variables
    !****************
    type(DataSetT)       :: oswDataSet, listDataSet ! their associated handles 
    type(ArrayT)         :: oswArray                ! the images
    type(TableT)         :: listTable               ! source list handle
    type(ColumnT)        :: pcolumn                 ! pointer to a column
 
    ! input Poisson stats or not

    integer (kind=int32)              :: boxScale                ! sliding box scale
    integer (kind=int32)             :: maxScale                ! Maximum scale to search
    logical (kind=bool)   :: poisson

    !***********************
    ! binning variables
    !***********************
    logical                      :: twixel

    character(len=6)            :: omFilter   ! the OM filter

    character(len=8)            :: omMode     ! FAST or IMAGING mode 

    !**************************************
    ! variables for the output sourcelist
    !**************************************
    integer  (kind=int32):: size = 5
   
    real (kind=single), dimension(:), allocatable :: xsource             ! x source ps (pix)
    real (kind=single), dimension(:), allocatable :: ysource             ! y source pos (px)
    real (kind=single), dimension(:), allocatable :: sourcesignif        ! source significance
    real (kind=single), dimension(:), allocatable :: countsource         ! source counts
    real (kind=single), dimension(:), allocatable :: counterr            ! error on source counts
    real (kind=single), dimension(:), allocatable :: sourceBackGroundCount  ! Source background counts
    real (kind=single), dimension(:), allocatable :: psfwidth_x          ! width of psf
    real (kind=single), dimension(:), allocatable :: psfwidth_y          ! width of psf
    real (kind=single), dimension(:), allocatable :: poserr              ! error on position
    real (kind=single), dimension(:), allocatable :: paSource            ! The position-angle for each source 
    real (kind=single), dimension(:), allocatable :: sourceApertureRadius      ! The aperture radii for each source
   
   
    integer (kind=int32), dimension(:), allocatable :: noSourcePixels ! The number of pixels included by a source
 
  
    integer (kind=int8), dimension(:), allocatable :: extendSource    ! Point-source (0), otherwise extended (1)
    integer (kind=int8), dimension(:), allocatable :: confuseSource   !
    integer (kind=int8), dimension(:), allocatable :: qualitySource   ! Set to 1 if 1 or more source pixels is bad
 
  

    integer (kind=int32), pointer :: srcnum(:)  ! the source number
    real (kind=single), pointer   :: r32(:)     ! pointer for list
    real (kind=double), pointer   :: r64(:)     ! pointer for list
 
    !***********************
    ! input image variables
    !***********************

    integer (kind=int8)                      :: ndim     ! no of dimensions
    integer (kind=int32)                     :: nx, ny   ! x,ysize of image
    integer (kind=int32)                     :: dim(2)   ! image dimension array
    real (kind=single)                       :: exposure
    real (kind=single), pointer   :: image(:, :)
    integer (kind=int16), pointer :: int16array(:, :)

  
    !************************************
    ! Background determination variables
    !************************************
    real (kind=single)               :: smoothSize   ! background smoothing size
    real (kind=single)               :: minDim

    !******************************
    ! source searching variables
    !******************************
    integer (kind=int32)              :: nSource
    real (kind=single), allocatable     :: background(:,:)
    real (kind=single), allocatable     :: smBackground(:, :)
    real (kind=single)                  :: medianBackground
    integer (kind=int32)              :: i, j, m, k, l        ! loop counters
    integer (kind=int32)              :: lPlus1
    real (kind=single)                  :: nSigma
    real (kind=single)                  :: fraction
    real (kind=single)                  :: minSigma
    real (kind=single)                  :: width
    real (kind=single)                  :: psfSemiA             ! The computed semi-major axis of the psf map
    real (kind=single)                  :: psfSemiB             ! The computed semi-minor axis of the psf map
    real (kind=single)                  :: maxPSFWidth          ! The maximum of the above two
    character(len=1024)   :: outputBoxImage
    character(len=1024)   :: significanceMap

    integer(kind=int16), allocatable  :: levelImage(:, :)
    integer (kind=int32)              :: allocErr
    integer (kind=int32)              :: nSources
    integer (kind=int32)              :: sourceConnect

    type(sourceParams), pointer :: sourcePars(:)
    integer (kind=int32)              :: nSourcePars
    integer,allocatable :: minX(:), maxX(:), minY(:), maxY(:)
    logical (kind=bool)  :: outputRegionFile
    integer  (kind=int32) :: ioUnit
    real(kind=double) :: plateScale, lambda

    real (kind=single) :: detectionRate
    real (kind=single) :: nullValue
    real (kind=double) :: apertureRadius     ! Radius (pixels) of circular aperture to be used for photometry
    real (kind=double) :: distance, tolerance ! distance and its minimal permitted value for the detected sources
!
!-------------------------------------------------------------------------------------
!
    type(TableT)                                     :: inTableH     ! Input table handle.
    type(ColumnT), dimension(4)                      :: inColumnH    ! Input image column handle.

    integer (kind=int32)                                          :: istat
 
    integer(kind=int32)                              :: windowX0     ! WindowX0.
    integer(kind=int32)                              :: windowY0     ! WindowY0.
    integer(kind=int32)                              :: windowDX     ! WindowDX.
    integer(kind=int32)                              :: windowDY     ! WindowDY.

    integer(kind=int32), dimension(:), pointer       :: windowX0P    ! WindowX0 pointer.
    integer(kind=int32), dimension(:), pointer       :: windowY0P    ! WindowY0 pointer.
    integer(kind=int32), dimension(:), pointer       :: windowDXP    ! WindowDX pointer.
    integer(kind=int32), dimension(:), pointer       :: windowDYP    ! WindowDY pointer.

    integer(kind=int32)                              :: xMin         ! Minimum x value for window.
    integer(kind=int32)                              :: xMax         ! Maximum x value for window.
    integer(kind=int32)                              :: yMin         ! Minimum y value for window.
    integer(kind=int32)                              :: yMax         ! Maximum y value for window.

    type(ArrayT)                                     :: qualArrayH   ! Quality Array handle.
    integer(kind=int8), dimension(:, :), pointer     :: qualArrayP   ! Quality array pointer.
    logical(kind=bool), dimension(:, :), allocatable :: qualArray    ! Quality array.
    logical(kind=bool)                               :: fastMode = .false.
    integer(kind=int32)                              :: filterIdInt
 
    nullValue = -real(99999)
  

    !*************************
    ! get the task parameters
    !*************************
    minSigma = 2

    set = stringParameter("set")
    outset = stringParameter("outset")
    nSigma = realParameter("nsigma")
    fraction = realParameter("contrast")
    backgroundImage = stringParameter("background")
    significanceMap = stringParameter("signifimage")
    levelImageName = stringParameter("levelimage")
    smoothSize = realParameter("smoothsize")
    boxScale = intParameter("boxscale")  ! Now should be entered in arcsecs (12.12.2000)
                                    
    maxScale = intParameter("maxscale")
    maxScale = MAX(maxScale, 1)
    sourceConnect = intParameter("pixelconnect")
    outputBoxImage = stringParameter("boximage")
    flatField = stringParameter("flatset")
    mod8noise = stringParameter("mod8set")
    regionfile = stringParameter("regionfile")
    outputRegionFile = booleanParameter("outputregionfile")
 
    !*************************
    ! get handle on osw image
    !*************************
    oswDataSet = DataSet(set,read)

    oswArray = array(oswDataSet, "PRIMARY", read)

    ndim  = numberOfDimensions(oswArray)             ! How many dimensions?

    ! If it's not an image then crash out
    if (ndim .ne. 2) call error('notImage','input file not an image')

    ! Get the exposure time in secs
    exposure = real32Attribute(oswDataSet,"EXPOSURE")

    ! Get the OM mode (IMAGING or FAST)
    omMode = stringAttribute(oswDataSet, "DATAMODE")
    if (trim(omMode) .eq. "FAST") then
        fastMode = .true.
    else
        fastMode = .false.
    end if 
         
    dim = dimensions(oswArray)               ! Get the size of the image

    nx = dim(1)
    ny = dim(2)

    !**********************************************************************
    ! convert the image (useful for testing on DSS images, which are int16)
    !**********************************************************************
    if (arrayDataType(oswArray) .eq. Real32) then
        image => real32Array2Data(oswArray)

    else
        int16array => int16array2data(oswArray)
        allocate(image(0 : nx - 1, 0 : ny - 1), stat = allocErr)
        call CheckAllocation(allocErr, "image")
        image = real(int16array)
    end if



    !**********************************
    ! compute binning
    !**********************************
    binx = int32Attribute(oswDataSet, "BINAX1")
    biny = int32Attribute(oswDataSet, "BINAX2")
    twixel = booleanAttribute(oswDataSet, "BINBPE")

    binx = 2 ** binx
    biny = 2 ** biny

    if (twixel) then
        binx = binx * 2
        biny = biny * 2
    end if

    if (binx .ne. biny) then
        call fatal("diffBin","Version 1 cannot deal with different binning factors in x and y")
    endif

    omFilter = stringAttribute(oswDataSet, "FILTER")


    !**************************************************
    ! Get a handle on the modes binary table extension.
    !**************************************************
    inTableH = table(oswDataSet, "MODES")

    !**************************************************
    ! Get handles and pointers on the modes binary table 
    ! extension columns and extract the window data.
    !**************************************************
    inColumnH(1) = column(inTableH, "WINDOWX0", read)
    windowX0P => int32Data(inColumnH(1))
    windowX0 = windowX0P(0)

    inColumnH(2) = column(inTableH, "WINDOWY0", read)
    windowY0P => int32Data(inColumnH(2))
    windowY0 = windowY0P(0)

    inColumnH(3) = column(inTableH, "WINDOWDX", read)
    windowDXP => int32Data(inColumnH(3))
    windowDX = windowDXP(0)

    inColumnH(4) = column(inTableH, "WINDOWDY", read)
    windowDYP => int32Data(inColumnH(4))
    windowDY = windowDYP(0)

    xMin = windowX0
    xMax = windowX0 + windowDx - 1
    yMin = windowY0
    yMax = windowY0 + windowDy - 1
 
    !**************************************
    ! Set the quality array.
    !**************************************
    allocate (qualArray(0 : nx - 1, 0 : ny - 1), stat = allocErr)
    call CheckAllocation(allocErr, "qualArray")

   
    ! Get a handle and pointer on the quality array and extract the quality data.
    qualArrayH = array(oswDataSet, "QUALITY", read)
    qualArrayP => int8Array2Data(qualArrayH)
  
    where (qualArrayP .ne. 0) 
        qualArray = .true.
    elsewhere
        qualArray = .false.      
    end where

    !*********************************
    !  Get the filter Id integer value
    !*********************************
    filterIdInt = filterIdToInt(omFilter)

    !**********************
    ! Set the CAL state
    !**********************
    call CAL_setState(OM, filterId = filterIdInt)

    
    !************************************* 
    ! Get the Plate scale (arcsecs/pixel)
    !*************************************
    call CAL_omGetPlateScale(lambda, plateScale)

    !****************************************
    ! Convert boxscale from arcsecs to pixels  
    !****************************************
    boxScale = ceiling(real(boxScale) / plateScale)


    !*************************************************************
    ! Set boxscale to take into account PSF of different filters
    ! 5 pixels minimum for optical filters (white,V,B FWHM ~ 2.6 pixels) 
    ! 7 pixels minimum for U+UV filters  (U,UV* FWHM ~ 3.8 pixels)
    !*************************************************************

    select case (omFilter)
        case ("V", "B", "WHITE", "white", "MAGNIF")
            boxScale = max(boxScale, 5)

        case ("U", "UVW1", "UVW2", "UVM2")
            boxScale = max(boxScale, 7)

         case default
             call fatal("WrongFilter","Filter in not one of White,WHITE,V,B,U,UVW1,UVW2,UVM2, or MAGNIF")

    end select

    !*********************************************************
    ! Take into account binning but make sure at least 3x3 box
    !*********************************************************
    boxScale = max(boxScale / binx, 3)

    write(errstr,'(a, i3)') "Minimum detection box set to : ", boxScale
    call message(AppMsg, Verbose)

    !************************************
    ! First smooth image on some scale 
    !************************************
    allocate(smBackground(0 : nx - 1, 0 : ny - 1), stat = allocErr)
    call CheckAllocation(allocErr, "smBackground")


    write(errstr,'(a)') "Beginning smoothing image (get median)"
    call message(AppMsg, Verbose)

    !********************************************************
    ! Smooth on small scale to get median value
    ! Needed if in poisson limit or low counts (almost always)
    !*********************************************************
    minDim = min(nx, ny) / 10.0

    smoothSize = min(smoothSize, minDim)
    smoothSize = max(smoothSize, 1.0)

    !********************************************
    ! image is the input image and smBackground
    ! the smoothed background image.
    !********************************************
    call smoothImage(image, nx, ny, smBackground, 3.0)

   
    write(errstr,'(a)') "Finished smoothing image "
    call message(AppMsg, Verbose)


    !**********************************
    ! get median of whole image
    !**********************************
    write(errstr,'(a)') "Getting medianBackground "
    call message(AppMsg, Verbose)

    call computeMode(smBackground, nx, ny, 3.0, medianBackground)

    write(errstr,'(a)') "Finished getting medianBackground "
    call message(AppMsg, Verbose)

    write(errstr,'(a, f7.1)') "Median of image: ", medianBackground
    call message(AppMsg, Verbose)

    if(medianBackground * boxScale * boxscale .lt. 15) then
        write(errstr,'(a, f7.1)') "Assuming Poisson statistics "
        call message(AppMsg, Verbose)
        poisson = .true.
    else
        write(errstr,'(a, f7.1)') "Assuming Gaussian statistics "
        call message(AppMsg, Verbose)
        poisson= .false.
    end if

    !***************************************************
    ! compute the background using a median box filter
    !***************************************************
    write(errstr,'(a)') "Computing the background"
    call message(AppMsg, Verbose)

    allocate(background(0 : nx - 1, 0 : ny - 1), stat = allocErr)
    call CheckAllocation(allocErr, "background")

    call computeBackground(image, nx, ny, background, smBackground, &
           medianBackground, 3.0)
  
    write(errstr,'(a)') "Beginning smoothing image (second time)"
    call message(AppMsg, Verbose)

    !**********************************************
    ! Make sure smooth size is less than image size
    !**********************************************

    minDim = min(nx, ny) / 10.0
    smoothSize = min(smoothSize, minDim)
    smoothSize = max(smoothSize, 1.0)


    call smoothImage(background, nx, ny, smBackground, smoothSize)
    write(errstr,'(a)') "Finished smoothing image (second time)"
    call message(AppMsg, Verbose)


    deallocate (background, stat = allocErr)
    call CheckDeallocation(allocErr, "background")

    write(errstr,'(a)') "Finished computing the background"
    call message(AppMsg, Verbose)


    !***********************************************
    ! Get numbers of detection limits
    !**********************************************
    call computeMode(smBackground, nx, ny, 3.0, medianBackground)

    write(errstr,'(a, f7.1)') "Smoothed median background : ", medianBackground
    call message(AppMsg, Verbose)
            
    !*****************************************
    ! Output background image if wanted
    !***************************************** 
    if(backgroundImage .ne. "")  call SaveBackgroundImage(smBackground, nx, ny, backgroundImage, "Background")


    !***************************
    ! Get significance Map
    !***************************

    write(errstr,'(a)') "Running source detection algorithm"
    call message(AppMsg, Verbose)

    allocate(levelImage(0 : nx - 1, 0 : ny - 1), stat = allocErr)
    call CheckAllocation(allocErr, "levelImage")

    levelImage = 0
    !*****************************************************
    ! Find sources in the image (image is the input image 
    ! and smBackground the smoothed image.
    ! levelImage is an image showing the initially 
    ! detected sources (Each pixel of a source has the
    ! pixel value set to the source number
    !*****************************************************
    call sourceDetectManyScales(image, nx, ny, smBackground, nSigma, minSigma, &
         boxScale, maxScale, nSources, levelImage, significanceMap, poisson)


    write(errstr,'(a,i5)') &
             "Finished running source detection algorithm - nsources: ",nSources

    call message(AppMsg, Verbose)

    write(errstr,'(''Nsources = '', i5)') nSources
    call message(AppMsg, Verbose)


    !*******************************************************
    ! Level image has all the information we need...
    ! Write this image to a file if requested
    !*******************************************************
    if(levelImageName .ne. "")  call SaveLevelImage(levelImage, nx, ny, levelImageName, "Level image")


    !*******************************************************
    ! Get min/max values for all sources from levelImage
    !******************************************************
    allocate(minX(nSources), stat = allocErr)
    call CheckAllocation(allocErr, "minX")

    allocate(maxX(nSources), stat = allocErr)
    call CheckAllocation(allocErr, "maxX")

    allocate(minY(nSources), stat = allocErr)
    call CheckAllocation(allocErr, "minY")

    allocate(maxY(nSources), stat = allocErr)
    call CheckAllocation(allocErr, "maxY")

    call getMinMaxFromLevelImage(levelImage, nx, ny, nSources, minX, maxX, minY, maxY)


    !*******************************************************
    ! Allocate memory for arrays that will store the source 
    ! parameters for writing to the output fits file
    !*******************************************************
    size = max(100, nSources * size) 
    allocate(xsource(1 : size), stat = allocErr)
    call CheckAllocation(allocErr, "xsource")

    allocate(ysource(1 : size), stat = allocErr)
    call CheckAllocation(allocErr, "ysource")

    allocate(sourcesignif(1 : size), stat = allocErr)
    call CheckAllocation(allocErr, "sourcesignif")

    allocate(countsource(1 : size), stat = allocErr)
    call CheckAllocation(allocErr, "countsource")

    allocate(counterr(1 : size), stat = allocErr)
    call CheckAllocation(allocErr, "counterr")
  
    allocate(sourceBackGroundCount(1 : size), stat = allocErr)
    call CheckAllocation(allocErr, "sourceBackGroundCount")
 
    allocate(sourceApertureRadius(1 : size), stat = allocErr)
    call CheckAllocation(allocErr, "sourceApertureRadius")
 

    allocate(psfwidth_x(1 : size), stat = allocErr)
    call CheckAllocation(allocErr, "psfwidth_x")

    allocate(psfwidth_y(1 : size), stat = allocErr)
    call CheckAllocation(allocErr, "psfwidth_y")

    allocate(poserr(1 : size), stat = allocErr)
    call CheckAllocation(allocErr, "poserr")
    
    allocate(paSource(1 : size), stat = allocErr)
    call CheckAllocation(allocErr, "paSource")

    allocate(extendSource(1 : size), stat = allocErr)
    call CheckAllocation(allocErr, "extendSource")

    allocate(confuseSource(1 : size), stat = allocErr)
    call CheckAllocation(allocErr, "confuseSource")

    allocate(qualitySource(1 : size), stat = allocErr)
    call CheckAllocation(allocErr, "qualitySource")


    allocate(noSourcePixels(1 : size), stat = allocErr)
    call CheckAllocation(allocErr, "noSourcePixels")


    !*******************************************************
    ! 6.6.2001 CB
    ! Fit a 2-d gaussian to the cal psf-image.
    ! These are needed
    ! for test to ascertain if a source is extended or not.
    !********************************************************


!--- In order to avoid problem with the empty field for the MAGNIFIER
!    the following routine (ComputePSFWidths) is omitted for nSources=0
!    (VNY 09.07.2001)
    if (nSources .gt. 0) then 
       call ComputePSFWidths(psfSemiA, psfSemiB)
       maxPSFWidth = 1.5 * max(psfSemiA, psfSemiB) / binX
    else
       maxPSFWidth = 0.0
    end if


    !************************************************************
    ! for each of the sources get sources even if blended
    !************************************************************
    k = 0


    do i = 1, nSources
        write(errstr, '(''Analysing source '', i5)') i
        call message(AppMsg, Verbose)
 
        call detectParams(nx, ny, image, smBackground, levelImage, i, &
             minX(i), maxX(i), minY(i), maxY(i), nSigma, fraction, &
             nSourcePars, sourcePars)
        
    
        !*********************************************************
        !---NB (VNY 10.09.2001): X and Y coordinates are in  the 
        ! CCD-coordinate system starting with (1,1)
        ! in the center of the left down corner pixel
        !*********************************************************
        if (fastMode) then
            call  FastChainChecks(minX(i), maxX(i), minY(i), maxX(i), &
                                      nx, ny, image, sourcePars(1))

        end if

        if(nSourcePars .gt. 0) then

            !**************************************
            ! remove identical sources if any
            !**************************************
            j = 0
            removeMultiple: do
                j = j + 1
                if(j .gt. nSourcePars) exit removeMultiple
                m = j
                removeMultiple2: do 
                    m = m + 1
                    if(m .gt. nSourcePars) exit removeMultiple2

                    !---VNY 30.05.2001 Check the distance between two source
                    distance=(sourcePars(m)%x - sourcePars(j)%x) ** 2 + &
                             (sourcePars(m)%y - sourcePars(j)%y) ** 2
                    distance = sqrt(distance)

                    tolerance = 1.0

                    ! VNY 07.08.2001 tolerance is set as maximum of two sources semi major axis
                    ! for the case of FAST mode
                    if (fastMode) then 
                       tolerance = max(sourcePars(j)%semiMajor, sourcePars(m)%semimajor)
                       tolerance = tolerance * 1.5
                    end if


                    if (distance .lt. tolerance) then

                        forall (l = m : nSourcePars - 1)
                            lPlus1 = l + 1
                            sourcePars(l) = sourceParams( &
                                        sourcePars(lPlus1)%x, sourcePars(lPlus1)%y, &
                                        sourcePars(lPlus1)%errX, sourcePars(lPlus1)%errY, &
                                        sourcePars(lPlus1)%PosAngle, sourcePars(lPlus1)%semiMajor, &
                                        sourcePars(lPlus1)%semiMinor, sourcePars(lPlus1)%countrate, &
                                        sourcePars(lPlus1)%countrateErr, sourcePars(lPlus1)%backGroundRate, &
                                        sourcePars(lPlus1)%noSourcePixels, sourcePars(lPlus1)%signif, &
                                        sourcePars(lPlus1)%Cflag)
                        end forall

                        nSourcePars = nSourcePars - 1
                        m = m - 1
                    end if 

                end do removeMultiple2
            end do removeMultiple

            do j = 1, nSourcePars
                if(sourcePars(j)%countrate .gt. 0) then
                    k = k + 1
                    xSource(k) = sourcePars(j)%x
                    ySource(k) = sourcePars(j)%y

                    !********************************************************
                    ! Set lower bound to 0.5 pixel for positional accuracy
                    !********************************************************

                    posErr(k) = max(max(sourcePars(j)%errX, sourcePars(j)%errY), 0.5) * &
                                max(binx, biny) * plateScale
                    paSource(k) = sourcePars(j)%PosAngle
                    countSource(k) = sourcePars(j)%countrate
                    countErr(k) = sourcePars(j)%countrateErr
                    sourceBackGroundCount(k) = sourcePars(j)%backGroundRate
                    psfwidth_x(k) = sourcePars(j)%semiMajor ! * binx * plateScale
                    psfwidth_y(k) = sourcePars(j)%semiMinor ! * biny * plateScale
                    sourceSignif(k) = sourcePars(j)%signif      
                    noSourcePixels(k) = sourcePars(j)%noSourcePixels

                    !********************
                    ! Set up flags  psf width in sigma (arc secs)
                    !********************
                    width = max(psfwidth_x(k), psfwidth_y(k))
                    !********************************************************
                    ! extended source flag if max width > 2 this is in sigma
                    ! 7th June 2001 psfWidth below replaces Jon's original 3 
                    !********************************************************
                    if (width .gt. maxPsfWidth) then  
                        extendSource(k) = 1
                    else
                        extendSource(k) = 0
                    end if
            
                    confuseSource(k) = sourcePars(j)%Cflag            
                end if
            end do ! j=1, nSourcePars

        endif
    end do   ! End of loop through all the sources

    nSource = k
    deallocate(nosourcePixels, stat = allocErr)
    call CheckDeallocation(allocErr, "noSourcePixels")

    deallocate(levelImage, stat = istat)
    call CheckDeallocation(istat, "levelImage")

    deallocate(minX, stat = allocErr)
    call CheckDeallocation(allocErr, "minX")

    deallocate(maxX, stat = allocErr)
    call CheckDeallocation(allocErr, "maxX")
 
    deallocate(minY, stat = allocErr)
    call CheckDeallocation(allocErr, "minY")

    deallocate(maxY, stat = allocErr)
    call CheckDeallocation(allocErr, "maxY")
  
    !**********************************
    ! Output to saoiamge region file
    !**********************************
    if(outputRegionFile) then
        call get_free_io_unit(ioUnit)
        open(unit = ioUnit, file = regionfile, status = 'unknown')
        do i = 1, nSource
            write(ioUnit, '('' ELLIPSE('', f7.2,'', '', f7.2, '', '', f6.2, '', '', f6.2, '', '', f6.2, '')'')') &
                  xSource(i), ySource(i), psfwidth_x(i) * 3, &    ! / (binx * plateScale), &
                  psfwidth_y(i) * 3, &   ! / (biny * plateScale), 
                  paSource(i)
        end do
        close(ioUnit)
    end if


    !*********************************************
    ! Write out the source list to the oswDataSet
    !*********************************************
    write(errstr,'(a)') "Writing source list"
    call message(AppMsg, Verbose)

    !**********************************
    ! Create the output source list
    !**********************************
    listDataSet = dataSet(outset, create)

    listTable = addTable(listDataSet, "SRCLIST", nsource)

    !****************************************
    ! copy the keywords from the image header
    !****************************************

    call copyAllAttributes(attributable(oswDataSet), attributable(listDataSet))

    !*****************************
    ! Add in detection count rate
    ! which will be stored in the source list file
    ! and used by ommag to calculate the
    ! background limiting  magnitude
    !************************************



!--- VNY 09.06.2001 In the case of no sources detected the following procedure will not be called
    if (nSources .gt. 0) then
       apertureRadius = 12.0 / real(binx)  
    else
       apertureRadius = 1.0
    end if


    if (medianBackground .gt. 0.0) then  ! was boxScale * boxScale- now the aperture area
        detectionRate = nSigma * sqrt(medianBackground * PI * apertureRadius * apertureRadius ) / exposure
    else
        detectionRate = 0.0
    end if 

    write(errstr,'(a, 1pg9.3)') "Detection Count Rate: ", detectionRate
    call message(AppMsg, Verbose)

    call setAttribute(listDataSet, "CRLIMIT", detectionRate, "counts/sec", "Detection limit count rate")

   
    ! this is an osw source list

    call setAttribute(listDataSet, "CONTENT", "OM OSW SOURCE LIST", &
                  "File contents")
    call setAttribute(listDataSet, "HDUCLASS", "OGIP", &
                  "Format conforms to OGIP/GSFC conventions")
    call setAttribute(listDataSet, "HDUCLAS1", "SRCLIST", &
                  "File contains a source list")

    !********************************
    ! now add the columns one by one
    ! starting with the source number
    !*********************************
    pcolumn = addColumn(listTable, "SRCNUM", Integer32, " ")
    srcnum => int32data(pcolumn)

    forall (i = 0 : nsource - 1) srcnum(i) = i +1
   
    call release(pcolumn)

    !*********************************
    ! Add the source positions
    !*********************************

    call AddReal32Column(listTable, "XPOS", "pixels", nsource, xsource, nullValue)
  
    call AddReal32Column(listTable, "YPOS", "pixels", nsource, ysource, nullValue)
   
    pcolumn = addColumn(listTable, "RA", Real64, "degrees")
    r64 => real64data(pcolumn)
    forall (i = 0 : nsource - 1) r64(i) = 0.0

    call release(pcolumn)

    pcolumn = addColumn(listTable, "DEC", Real64, "degrees")
    r64 => real64data(pcolumn)
    forall (i = 0 : nsource - 1) r64(i) = 0.0
    

    call release(pcolumn)

    pcolumn = addColumn(listTable, "LII", Real64, "degrees")
    r64 => real64data(pcolumn)
    forall (i = 0 : nsource - 1) r64(i) = 0.0
    call release(pcolumn)
    pcolumn = addColumn(listTable, "BII", Real64, "degrees")
    r64 => real64data(pcolumn)
    forall (i = 0 : nsource - 1) r64(i) = 0.0
    call release(pcolumn)




    call message(AppMsg, Verbose)

    call AddReal32Column(listTable, "POSERR", "pixels", nsource, poserr, nullValue)
    deallocate(poserr, stat = allocErr)
    call CheckDeallocation(allocErr, "poserr")
 

    !********************************************************************************
    ! In order to convert the source counts to magnitudes have to use a 12 pixel
    ! aperture radius to sum the counts
    ! subroutine AperturePhotometry conducts aperture photometry on each source
    !
    !*******************************************************************************
    call AperturePhotometry(nx, ny, image, qualArray, & 
                            real(apertureRadius), nsource, & 
                            xsource, ysource, psfwidth_x, psfwidth_y, & 
                            countSource, counterr, sourcesignif, &
                            sourceBackGroundCount, extendSource, confuseSource, &
                            qualitySource, sourceApertureRadius)

    deallocate(qualArray, stat = istat)
    call CheckDeallocation(istat, "qualArray")
  
    deallocate(xsource, stat = allocErr)
    call CheckDeallocation(allocErr, "xsource")

    deallocate(ysource, stat = allocErr)
    call CheckDeallocation(allocErr, "ysource")


    !*********************************************
    ! Add the source-count rates
    !*********************************************

    countsource = countsource / exposure
    call AddReal32Column(listTable, "RATE", "counts/sec", nsource, countsource, nullValue)
    deallocate(countsource, stat = allocErr)
    call CheckDeallocation(allocErr, "countsource")

    counterr = counterr / exposure
    call AddReal32Column(listTable, "RATE_ERR", "counts/sec", nsource, counterr, nullValue)
    deallocate(counterr, stat = allocErr)
    call CheckDeallocation(allocErr, "counterr")
  

    !***********************************************************************
    ! Add the aperture radius (06.06.2001 CB)
    !***********************************************************************
    call AddReal32Column(listTable, "APERTURE", "pixels", nsource, &
                         sourceApertureRadius, nullValue)
    deallocate(sourceApertureRadius, stat = allocErr)
    call CheckDeallocation(allocErr, "apertureRadius")

 
    !*************************************
    ! Add the source background rate
    !************************************
    sourceBackGroundCount = sourceBackGroundCount / exposure
    call AddReal32Column(listTable, "BACKGROUND_RATE", "counts/sec", &
                         nsource, sourceBackGroundCount, nullValue)
    deallocate(sourceBackGroundCount, stat = allocErr)
    call CheckDeallocation(allocErr, "sourceBackGroundCount")

    call AddReal32Column(listTable, "SIGNIFICANCE", "sigma", nsource, sourcesignif, nullValue)
 
    !************************************
    ! Add the magnitudes (set to zero!)
    !************************************ 
    sourcesignif = 0.0
    call AddReal32Column(listTable, "MAG", "magnitudes", nsource, sourcesignif, nullValue)
    call AddReal32Column(listTable, "MAGERR", "magnitudes", nsource, sourcesignif, nullValue)
    deallocate(sourcesignif, stat = allocErr)
    call CheckDeallocation(allocErr, "sourcesignif")

    pcolumn = addColumn(listTable, "FWHM_MAJ", Real32, "arcsecs")
    r32 => real32data(pcolumn)

    !*********************************************************
    ! Convert widths to arc secs
    !*********************************************************
    psfwidth_x = psfwidth_x * binx * plateScale
    psfwidth_y = psfwidth_y * biny * plateScale
    forall (i = 0 : nsource - 1) r32(i) = max(2.354 * psfwidth_x(i + 1), 2.354 * psfwidth_y(i + 1))
    
    call release(pcolumn)
  
    pcolumn = addColumn(listTable, "FWHM_MIN", Real32, "arcsecs")
    r32 => real32data(pcolumn)

    forall (i = 0 : nsource - 1) r32(i) = min(psfwidth_x(i + 1) * 2.354, psfwidth_y(i + 1) * 2.354)
  
    call release(pcolumn)
 
    deallocate(psfwidth_x, stat = allocErr)
    call CheckDeallocation(allocErr, "psfwidth_x")
    deallocate(psfwidth_y, stat = allocErr)
    call CheckDeallocation(allocErr, "psfwidth_y")

    call AddReal32Column(listTable, "PA", "degrees", nsource, paSource, nullValue)
    deallocate(paSource,  stat = allocErr)
    call CheckdeAllocation(allocErr, "paSource")

  
    !**********************************
    ! Add the quality flags
    !**********************************
    call AddInt8Column(listTable, "QFLAG", "none", nsource, qualitySource)   ! Can be either 0 or 1
    deallocate(qualitySource, stat = istat)
    call CheckDeallocation(istat, "qualitySource")

    call AddInt8Column(listTable, "CFLAG", "none", nsource, confuseSource)   ! Can be either 0 or 1
    deallocate(confuseSource,  stat = allocErr)
    call CheckdeAllocation(allocErr, "confuseSource")

  
    call AddInt8Column(listTable, "EFLAG", "none", nsource, extendSource)    ! Can be either 0 (point-source) or 1 (extende
    deallocate(extendSource,  stat = allocErr)
    call CheckdeAllocation(allocErr, "extendSource")

    !************************
    ! release the memory
    !************************
    call release(oswArray)
    call release(listTable)
    call release(listDataSet)
    call release(oswDataSet)

 
    write(errstr,'(a)') "Exiting..."
    call message(AppMsg, Verbose)

end subroutine omdetect

!******************************************************************
! Get min and max values for each source from levelImage
!
!******************************************************************
subroutine getMinMaxFromLevelImage(levelImage, nx, ny, nSources, &
           minX, maxX, minY, maxY)

    integer  (kind=int32), intent(in) :: nx, ny
    integer (kind=int16), intent(in) :: levelImage(0 : nx - 1, 0 : ny - 1)
    integer (kind=int32), intent(in) :: nSources
    integer (kind=int32), intent(out) :: minX(nSources)
    integer (kind=int32), intent(out) :: maxX(nSources)
    integer (kind=int32), intent(out) :: minY(nSources)
    integer (kind=int32), intent(out) :: maxY(nSources)

    integer (kind=int32) :: i, j
    integer (kind=int32) :: levelImageVal



    forall (i = 1 : nSources)
        minX(i) = nx + 1
        maxX(i) = 0
        minY(i) = ny + 1
        maxY(i) = 0
    end forall

    do i = 0, ny - 1
        do j = 0, nx - 1
            levelImageVal = levelImage(j, i)

            if(levelImageVal .ne. 0) then
                minX(levelImageVal) = MIN(minX(levelImageVal), j + 1)
                maxX(levelImageVal) = MAX(maxX(levelImageVal), j + 1)
                minY(levelImageVal) = MIN(minY(levelImageVal), i + 1)
                maxY(levelImageVal) = MAX(maxY(levelImageVal), i + 1)
            end if
        end do
    end do

end subroutine getMinMaxFromLevelImage
!***************************************************************
! subroutine nextToPosition
!
!***************************************************************
recursive subroutine nextToPosition(newImage, levelImage, nx, ny, xpos, ypos, ok)

    integer (kind=int32), intent(in) :: nx, ny
    real, intent(inout) :: newImage(nx, ny)
    real, intent(inout) :: levelImage(nx, ny)
    integer (kind=int32), intent(in) :: xpos, ypos
    logical, intent(out) :: ok

    integer (kind=int32) :: minXpos, maxXpos, minYpos, maxYpos
    integer (kind=int32) :: i, j

!  print *,"nextToPosition"
  
    ok = .false.

    minXpos = MAX(1, xpos - 1)
    minYpos = MAX(1, ypos - 1)
    maxXpos = MIN(nx, xpos + 1)
    maxYpos = MIN(ny, ypos + 1)
    do i = minYpos, maxYpos
        do j = minXpos, maxXpos
            if(levelImage(j, i) .ne. 0) then
                if(i .ne. xpos .and. j .ne. ypos) then
                    if(newImage(j, i) .gt. 0) then
                        levelImage(xpos, ypos) = levelImage(j, i)
                        newImage(xpos, ypos) = 0
                        call nextToPosition(newImage, levelImage, nx, ny, j, i, ok)
                        ok = .true.
                    end if
                end if
            end if
        end do
    end do


end subroutine nextToPosition

!************************************************************
! subroutine sourceDetectManyScales
! Purpose: Detects all the sources in the image "image"- some may be
! detected twice.
! Returns: The number of sources detected "nSources" and the image
! "levelImage" - 
! image is the input image
! nx & ny the image dimensions
! background is the smoothed background image
!************************************************************
subroutine sourceDetectManyScales(image, nx, ny, background, nSigma, minSigma, &
     boxScale, maxScale, nSources, levelImage, significanceMap, poisson)

    integer (kind=int32), intent(in) :: nx, ny
    real (kind=single), intent(in) :: image(0 : nx - 1, 0 : ny - 1)
    real (kind=single), intent(in) :: background(0 : nx - 1, 0 : ny - 1)
    real (kind=single), intent(in) :: nSigma, minSigma
    integer (kind=int32), intent(in) :: boxScale
    integer (kind=int32), intent(in) :: maxScale
    integer (kind=int32), intent(out) :: nSources
    integer (kind=int16), intent(out) :: levelImage(0 : nx - 1, 0 : ny - 1)
    character(len=*), intent(in) :: significanceMap
    logical(kind=bool), intent(in) :: poisson

    type(DataSetT)       :: significanceSet 
    type(ArrayT)         :: significanceArray         
    real (kind=single), dimension(: , :), pointer :: significanceP
  
    integer (kind=int32) :: origNx, origNy
    integer (kind=int32) :: oldNx, oldNy
    integer (kind=int32) :: newNx, newNy
    integer (kind=int32) :: nSource

    integer (kind=int32) :: allocErr
    real (kind=single), allocatable :: sigmaMap(:, :)
    real (kind=single), allocatable :: oldSigmaMap(:, :)
    real (kind=single), allocatable :: inImage(:, :)
    real (kind=single), allocatable :: inBackground(:, :)
    real (kind=single), allocatable :: newImage(:, :)
    real (kind=single), allocatable :: newBackground(:, :)

    real (kind=single):: sigmaMapThreshold
    real (kind=single):: minSigmaMapThreshold

    integer (kind=int32) :: scale

    !*************************************
    ! Do things at first scale 
    !*************************************
    allocate(oldSigmaMap(0 : nx - 1, 0 : ny - 1), stat = allocErr)
    call CheckAllocation(allocErr, "oldSigmaMap")

    ! Allocatable arrays of variable size - start at level 1

    !*************************************
    ! Do it at first scale
    !*************************************
    write(errstr, '(a)') "Running source detection algorithm (level 1)"
    call message(AppMsg, Verbose)  ! Verbose was 5

    call sourceDetect(image, nx, ny, background, oldSigmaMap, &
       sigmaMapThreshold, minSigmaMapThreshold, nSigma,minSigma, boxScale, &
       poisson)

    write(errstr,'(a)') "Finished running source detection algorithm (level 1)"
    call message(AppMsg, Verbose)  ! Verbose was 5

    !**********************************
    ! Output significance map if wanted
    !***********************************
    if(significanceMap .ne. "") then


        !**********************
        ! Create image file 
        !**********************
        significanceSet = DataSet(significanceMap, Create)

        significanceArray = addArray(significanceSet, "Significance Map", Real32, &
                            shape(oldSigmaMap))
     
        significanceP => real32Array2Data(significanceArray)
     
        significanceP = oldSigmaMap
     
        call release(significanceArray)
        call release(significanceSet)
     
    end if

    !**************************************
    ! Now start to make level image
    !**************************************
    write(errstr,'(a)') "Getting source number image (level 1)"
    call message(AppMsg, Verbose)  ! Verbose was 5

    nSource = 0

    allocate(sigmaMap(0 : nx - 1, 0 : ny - 1), stat = allocErr)
    call CheckAllocation(allocErr, "sigmaMap")
    sigmaMap = oldSigmaMap

    !print *,"SigmaMapThreshold:", sigmaMapThreshold
    call setSources(sigmaMap, nx, ny, 1, oldSigmaMap, levelImage, nx, ny, &
               sigmaMapThreshold, minSigmaMapThreshold, nSource)

    if(allocated(sigmaMap)) then
        deallocate(sigmaMap, stat = allocErr)
        call CheckDeallocation(allocErr, "sigmaMap")
    end if

    write(errstr, '(a)') "Finished getting source number image (level 1)"
    call message(AppMsg, Verbose)  ! Verbose was 5
 
    origNx = nx
    origNy = ny

    ! Now go down in levels (ie. rebinning of image)

    oldNx = nx
    oldNy = ny
    allocate(inImage(0 : oldNx - 1, 0 : oldNy - 1), stat = allocErr)
    call CheckAllocation(allocErr, "inImage")
    inImage = image
    allocate(inBackground(0 : oldNx - 1, 0 : oldNy - 1), stat = allocErr)
    call CheckAllocation(allocErr, "inBackground")


    inBackground = background

    scale = 1
    scaleLoop: do

        scale = scale + 1
        if(scale .gt. maxScale) exit scaleLoop

        write(errstr,'(a, i5)') "Analysing image at level ",scale
        call message(AppMsg, Verbose) ! Verbose was 5

        newNx = oldNx / 2
        newNy = oldNy / 2

        ! First - don't want to get tiny image!

        if(newNx .lt. 32 .or. newNy .lt. 32) exit scaleLoop
       

        allocate(newImage(0 : newNx - 1, 0 : newNy - 1), stat = allocErr)
        call CheckAllocation(allocErr, "newImage (sourceDetectManyScalee")

        allocate(newBackground(0 : newNx - 1, 0 : newNy - 1), stat = allocErr)
        call CheckAllocation(allocErr, "newBackground")

        !***********************************
        ! Rebin images by factor of 2
        !***********************************
        write(errstr, '(a, i5)') "Rebinning image at level ", scale
        call message(AppMsg, Verbose)  ! Verbose was 5

        call rebinImageBy2(inImage, oldNx, oldNy, newImage, newNx, newNy)
        call rebinImageBy2(inBackground, oldNx, oldNy, newBackground, newNx, newNy)

        ! Allocate sigmaMap

        allocate(sigmaMap(0 : newNx - 1, 0 : newNy - 1), stat = allocErr)
        call CheckAllocation(allocErr, "sigmaMap")
    
        !********************
        ! Source detect
        !********************


        write(errstr, '(a, i5)') "Source detect at level ", scale
        call message(AppMsg, Verbose)  ! Verbose was 5
        call sourceDetect(newImage, newNx, newNy, newBackground, sigmaMap, &
                          sigmaMapThreshold, minSigmaMapThreshold, nSigma, &
                          minSigma, boxScale, poisson)




        !*****************************
        ! add to levelImage map
        !*****************************
        call setSources(sigmaMap, newNx, newNy, scale, oldSigmaMap, levelImage, origNx, &
                        origNy, sigmaMapThreshold, minSigmaMapThreshold, nSource)


        if(allocated(sigmaMap)) then
            deallocate(sigmaMap, stat = allocErr)
            call CheckDeallocation(allocErr, "sigmaMap")
        end if

        write(errstr, '(a, i4)') "Finished getting source number image at scale ", scale
        call message(AppMsg, Verbose)  ! Verbose was 5


        ! Tidy up image arrays

        if(allocated(inImage)) then
            deallocate(inImage, stat = allocErr)
            call CheckDeallocation(allocErr, "inImage")
        end if

        if(allocated(inBackground)) then
            deallocate(inBackground,stat = allocErr)
            call CheckDeallocation(allocErr, "inBackground")
        end if

        allocate(inImage(0 : newNx - 1, 0 : newNy - 1), stat = allocErr)
        call CheckAllocation(allocErr, "inImage")

        inImage = newImage
        allocate(inBackground(0 : newNx - 1, 0 : newNy - 1), stat = allocErr)
        call CheckAllocation(allocErr, "inBackground (sourceDetectManyScales")

        inBackground = newBackground

        if(allocated(newimage)) then
            deallocate(newImage, stat = allocErr)
            call CheckDeallocation(allocErr, "newIMage")
       end if

        if(allocated(newBackground)) then
            deallocate(newBackground,stat=allocErr)
            call CheckAllocation(allocErr, "newBackground")
       end if

        oldNx = newNx
        oldNy = newNy

    end do scaleLoop

    if(allocated(inImage)) then
        deallocate(inImage, stat = allocErr)
        call CheckDeallocation(allocErr, "inImage")
    end if

    if(allocated(inBackground)) then
        deallocate(inBackground, stat = allocErr)
        call CheckDeallocation(allocErr, "inBackground")
    end if

    if(allocated(oldSigmaMap)) then
        deallocate(oldSigmaMap, stat = allocErr)
        call CheckDeallocation(allocErr, "oldSigmaMap")
    end if

    !***************************************************
    ! Get nSources from the maximum value  in levelImage
    !***************************************************
    nSources = MAXVAL(levelImage)
end subroutine sourceDetectManyScales

!****************************************************************
! subroutine rebinImageBy2 
!
!****************************************************************
subroutine rebinImageBy2(inImage, oldNx, oldNy, newImage, newNx, newNy)
 
    integer (kind=int32), intent(in) :: oldNx, oldNy
    real (kind=single), intent(in) :: inImage(0 : oldNx - 1, 0 : oldNy - 1)
    integer (kind=int32), intent(in) :: newNx, newNy
    real (kind=single), intent(out) :: newImage(0 : newNx - 1, 0 : newNy - 1)

    integer (kind=int32) :: i, j, k, l
    real (kind=single):: newImageVal

!  print *,"rebinImageBy2"

    if((oldNx / 2 .ne. newNx) .or. (oldNy / 2 .ne. newNy)) then
     call fatal("rebinImageBy2","Input arrays not factors of 2")
  endif

  forall (i = 0 : newNy - 1)
     k = i * 2
     forall (j = 0 : newNx - 1)
        l = j * 2
        newImageVal = inImage(l, k)
        newImageVal = newImageVal + inImage(l + 1, k)
        newImageVal = newImageVal + inImage(l, k + 1)
        newImageVal = newImageVal + inImage(l + 1, k + 1)
        newImage(j, i) = newImageVal
     end forall
  end forall

end subroutine rebinImageBy2
!****************************************************
! subroutine sourceDetect
! Sliding box algorithm for source detection
!****************************************************

subroutine sourceDetect(image, nx, ny, background, sigmaMap, &
     sigmaMapThreshold, minSigmaMapThreshold, nSigma, minThreshold, &
     boxScale, poisson)
 
    integer (kind=int32), intent(in) :: nx, ny
    real (kind=single), intent(in) :: image(0 : nx - 1, 0 : ny - 1)
    real (kind=single), intent(in) :: background(0 : nx - 1, 0 : ny - 1)
    real (kind=single), intent(inout) :: sigmaMap(0 : nx - 1, 0 : ny - 1)
    real (kind=single), intent(inout) :: sigmaMapThreshold
    real (kind=single), intent(inout) :: minSigmaMapThreshold
    real (kind=single), intent(in) :: nSigma
    real (kind=single), intent(in) :: minThreshold
    integer (kind=int32), intent(in) :: boxScale
    logical(kind=bool), intent(in) :: poisson

  ! local variables

    real (kind=single), allocatable   :: scan(:)
    integer (kind=int32), allocatable :: nscan(:)

    real (kind=single), allocatable :: newBAckground(:, :)
    real (kind=single), allocatable :: boxImage(:, :)

    integer (kind=int32) :: i, j
    integer (kind=int32) :: height
    real (kind=single):: backgroundVal, sigmaMapVal
    integer (kind=int32) :: imageVal
    integer (kind=int32) :: allocErr


    height = boxScale / 2

    allocate(scan(-height : height), stat = allocErr)
    call CheckAllocation(allocErr, "scan (sourceDetect)")


    allocate(nscan(-height : height), stat = allocErr)
    call CheckAllocation(allocErr, "nscan (sourceDetect)")

    allocate(boxImage(0 : nx - 1, 0 : ny - 1), stat = allocErr)
    call CheckAllocation(allocErr, "boxImage (sourceDetect)")

    allocate(newBackground(0 : nx - 1, 0 : ny - 1), stat = allocErr)
    call CheckAllocation(allocErr, "newBackground (sourceDetect)")

    !************************************     
    ! Slide the box to get binned image
    !************************************
    do i = 0, nx - 1
        call MovingBox(image, nx, ny, boxImage, height, height, i, scan, nscan, 2)
    end do

    !******************************************************
    ! Then background map (mod-8 noise on small scales!)
    !******************************************************

    do i = 0, nx - 1
        call MovingBox(background, nx, ny, newBackground, height, height, i, &
                       scan, nscan, 0)
    end do

    if(.not. poisson) then

        sigmaMapThreshold = nSigma
        minSigmaMapThreshold = minThreshold
     
        sigmaMap = (boxImage - newBackground) / sqrt(newBackground)

    else
        !******************************************************************
        ! Compute where data is above significance using Poisson statistics
        ! Calculate image as sigma not probability
        !******************************************************************
     
        sigmaMapThreshold = -log(newerrf(nSigma))
        minSigmaMapThreshold = -log(newerrf(minThreshold))

        do i = 0, ny - 1
            do j = 0, nx - 1
                backgroundVal = newBackground(j, i)
                imageVal = nint(boxImage(j, i))
           
                ! Get prob in log space
           
                if(imageVal .gt. backgroundVal) then
!---VNY 23.10.2001 Check backgroundVal and imageVal before calculation of log and lnFactrl
                   if (backgroundVal .gt. 0.0 .and. imageVal .gt. 0.0) then
                      sigmaMapVal = imageVal * log(backgroundVal) - backgroundVal - &
                                  lnFactrl(imageVal)
                   else
                      sigmaMapVal = 0.0
                   end if
                else
                    sigmaMapVal = 0.0
                end if
           
                sigmaMap(j, i) = -sigmaMapVal
            end do
        end do
      
    end if

    if(allocated(newBackground)) then
        deallocate(newBackground, stat = allocErr)
        call CheckDeallocation(allocErr, "newBackground")
    end if

    if(allocated(boxImage)) then
        deallocate(boxImage, stat = allocErr)
        call CheckDeallocation(allocErr, "boxImage")
    end if

    if(allocated(scan)) then
        deallocate(scan, stat = allocErr)
        call CheckDeallocation(allocErr, "scan")

    end if

    if(allocated(nscan)) then
        deallocate(nscan, stat = allocErr)
        call CheckDeallocation(allocErr, "nscan")
    end if

end subroutine sourceDetect

!***********************************************************************
!
! subroutine setSources - use 8 connectivity to match pixels
! fill in an image of where the sources are - the used to construct pixel
! lists to match output needed for TJH version of omdetect
!***********************************************************************
subroutine setSources(image, nx, ny, scale, fullImage, levelImage, oldNx, oldNy, &
     nSigma, minThreshold, nSources)

    !*********************************************************
    ! Sets up levelImage which contains only connected pixels
    ! Basically sets up where the sources are
    !*********************************************************
    integer  (kind=int32), intent(in) :: nx,ny
    real (kind=single), intent(inout) :: image(0 : nx - 1, 0 : ny - 1)
    integer  (kind=int32), intent(in) :: scale
    integer  (kind=int32), intent(in) :: oldNx, oldNy
    real (kind=single), intent(in) :: fullImage(0 : oldNx - 1, 0 : oldNy - 1)
    integer (kind=int16), intent(inout) :: levelImage(0 : oldNx - 1, 0 : oldNy - 1)
    real (kind=single), intent(in) :: nSigma
    real (kind=single), intent(in) :: minThreshold
    integer (kind=int32), intent(inout) :: nSources

    type(src), pointer :: sourceList
    integer (kind=int32) :: nSourceList

    integer (kind=int32) :: nPixel
    type(src), pointer :: currentSource
    type(srcpixel), pointer :: currentPixel
  
    integer (kind=int32) :: minXpos, maxXpos
    integer (kind=int32) :: minYpos, maxYpos

    integer (kind=int32) :: i, j
    integer (kind=int32) :: sourceNumber
    integer (kind=int32) :: xpos, ypos
    integer (kind=int32) :: numberSources
    integer (kind=int32) :: scaling

!  print *,"setSources"

    nullify(sourceList) ! source list for this scale
    nSourceList = 0
  
    !*****************************************************
    ! At this scale we get linked list of source positions
    !*****************************************************
    write(errstr,'(a, i4)') "Getting linked sources at scale ", scale
    call message(AppMsg, Verbose)  ! Verbose was 5

    call getLinkedSources(image, nx, ny, nSigma, minThreshold, sourceList, nSourceList)

    write(errstr,'(a, i4)') "Finished getting linked sources at scale ", scale
    call message(AppMsg, Verbose)  ! Verbose was 5

    if(nSourceList .eq. 0) then
        return
    end if
    !************************************************************************
    ! Check source lists with image/linkedImage to make sure new sources are 
    ! not repeats of sources found at a higher resolution
    !************************************************************************
    currentSource => sourceList
    sourceLoop: do
     
        if(.not. associated(currentSource)) then
        exit sourceLoop
     else
        sourceNumber = currentSource%sourceNum
        nPixel = currentSource%numPixels

        if(nPixel .ge. 0) then

           ! data stored

           if(scale .eq. 1) then  ! at minimum scale

              nSources = max(nSources, sourceNumber)
              currentPixel => currentSource%pixelList
              pixelLoop: do 
                 if(.not. associated(currentPixel)) then
                    exit pixelLoop
                 else
                    xpos = currentPixel%x
                    ypos = currentPixel%y
 

                    ! Not at multiple scales so just pixel/pixel

                    if(levelImage(xpos, ypos) .eq. 0) then
                       if(fullImage(xpos, ypos) .gt. minThreshold) then
                          levelImage(xpos, ypos) = sourceNumber
                       end if
                    end if
                 end if

                 ! Onto next pixel

                 currentPixel => currentPixel%next

              end do pixelLoop

           else

              numberSources = 0
              currentPixel => currentSource%pixelList
              pixelLoop2: do 
                 if(.not. associated(currentPixel)) then
                    exit pixelLoop2
                 else
                    xpos = currentPixel%x
                    ypos = currentPixel%y
                    
                    ! See if this source overlaps with other sources
                    
                    scaling = 2 ** (scale - 1)
                    minYpos = ypos * scaling
                    maxYpos = minYpos + scaling - 1
                    minXpos = xpos*scaling
                    maxXpos = minXpos + scaling - 1

                    !***********************************************
                    ! If there are sources already then ignore - 
                    ! already go down to minTresh at searching stage
                    !***********************************************
                    countSources: do i = minYpos, maxYpos
                       
                       do j = minXpos, maxXpos
                          
                          if(levelImage(j, i) .gt. 0)then
                             numberSources = -1
                             exit countSources
                          end if
                          
                       end do
                       
                    end do CountSources

                 end if

                 ! Onto next pixel

                 currentPixel => currentPixel%next

              end do pixelLoop2

              ! Now know if this source is clear of already detected sources
              ! If we don't have sources then assign pixels
                    
              if(numberSources .ge. 0) then ! not multiple sources
                       
                 ! If new sources then increment number of sources
                       
                 if(numberSources .eq. 0) then
                    nSources = nSources + 1
                    numberSources = nSources
                 end if
                 
                 currentPixel => currentSource%pixelList
                 pixelLoop3: do 
                    if(.not. associated(currentPixel))then
                       exit pixelLoop3
                    else
                       xpos = currentPixel%x
                       ypos = currentPixel%y


                             
                       ! See if this source overlaps with other sources
                    
                       scaling = 2 ** (scale - 1)
                       minYpos = ypos * scaling
                       maxYpos = minYpos + scaling - 1
                       minXpos = xpos * scaling
                       maxXpos = minXpos + scaling - 1
                       
                       
                       ! Now assign values and add pixels to linked list
                       
                       do i = minYpos, maxYpos
                          do j = minXpos, maxXpos
                             
                             ! Test to see if we've been here before
                             ! and if we're in a detected pixel possibility
                             
                             if(levelImage(j, i) .eq. 0) then
                                if(fullImage(j, i) .gt. minThreshold)then
                                   
                                   ! Is this connected to any other pixel?
                                   ! Use nearest neighbour again at this level
                                   ! within the range allowed by detection
                                   
                                   call nearestNeighbourImage(fullImage, &
                                        oldNx, oldNy, levelImage, minXpos, &
                                        maxXpos, minYpos, maxYpos, j, i, &
                                        minThreshold, numberSources)
                                   
                                end if
                             end if
                          end do
                          
                       end do

                       ! Onto next pixel

                       currentPixel => currentPixel%next

                    end if

                 end do pixelLoop3

              end if

           end if

        end if

        ! Onto next source

        currentSource => currentSource%next

     endif

  end do sourceLoop

  call freeSourceList(sourceList)

end subroutine setSources
!*****************************************
! subroutine freeSourceList
! Routine to free memory from sourceList
!*****************************************
subroutine freeSourceList(sourceList)

  type(src), pointer :: sourceList
  type(src), pointer :: currentSource
  type(srcPixel), pointer :: currentPixel

  integer (kind=int32)             :: allocErr

!  print *,"freeSourceList"

  currentSource => sourceList
  downSourceList: do

     if(associated(currentSource%next)) then
        currentSource => currentSource%next
     else
        exit downSourceList
     end if

  end do downSourceList

  upSourceList: do 

     currentPixel => currentSource%pixelList
     downPixelList: do
        if(associated(currentPixel%next)) then
           currentPixel => currentPixel%next
        else
           exit downPixelList
        end if
     end do downPixelList
     
     upPixelList: do
        if(associated(currentPixel%previous)) then
            currentPixel =>  currentPixel%previous
            deallocate(currentPixel%next,stat = allocErr)
            call CheckDeallocation(allocErr, "currentpixel%next")
            nullify(currentPixel%next)
        else
            exit upPixelList
        end if
     end do upPixelList
     deallocate(currentPixel, stat = allocErr)
     call CheckDeallocation(allocErr, "currentPixel")
          
     nullify(currentPixel)

     if(associated(currentSource%previous)) then
         currentSource => currentSource%previous
         deallocate(currentSource%next, stat = allocErr)
         call CheckDeallocation(allocErr, "currentSource%next")
         nullify(currentSource%next)
     else
         deallocate(currentSource, stat = allocErr)
         call CheckDeallocation(allocErr, "currentSource")
         exit upSourceList
     end if

  end do upSourceList

  nullify(sourceList)

end subroutine freeSourceList
!******************************************
! subroutine getLinkedSources
!
! Routine to get nearest neighbour 
!******************************************
subroutine getLinkedSources(image, nx, ny, nSigma, minSigma, sourceList, nSourceList)

    integer (kind=int32), intent(in) :: nx, ny
    real (kind=single), intent(inout) :: image(0 : nx - 1, 0 : ny - 1)
    real (kind=single), intent(in) :: nSigma, minSigma
    type(src), pointer :: sourceList
    integer (kind=int32), intent(inout) :: nSourceList

    type(src), pointer :: currentSource
    type(srcpixel), pointer :: currentPixel !, nextPixel
 
    integer (kind=int32) :: maxpos(2)
    integer (kind=int32) :: xpos, ypos
    integer (kind=int32) :: allocErr

    real (kind=single):: tmp

!  print *,"getLinkedSources"

    maxpos = MAXLOC(image)

    xpos = maxpos(1) - 1
    ypos = maxpos(2) - 1

    ! If no pixels satisfy criterion

    if(image(xpos, ypos) < nSigma) then
        return
    endif

    tmp = image(xpos, ypos)

    ! We have some pixels that satisfy criterion

    nSourceList = 1
    allocate(sourceList, stat = allocErr)
    call CheckAllocation(allocErr, "sourceList")

    sourceList%numPixels = 0
    nullify(sourceList%previous)
    currentSource => sourceList

    !************************************************
    ! Loop round discrete sources to get final list
    !************************************************
    linkSources: do

        currentSource%sourceNum = nSourceList
        allocate(currentSource%pixelList, stat = allocErr)
        call CheckAllocation(allocErr, "pixelList (getLinkedSources)")

        ! Set the number of pixels to 1
     
        currentSource%numPixels = 1
     
        ! Now add in pixel numbers
     
        currentSource%pixelList%x = xpos
        currentSource%pixelList%y = ypos
     
        nullify(currentSource%pixelList%next)
        nullify(currentSource%next)
     
        image(xpos, ypos) = 0  ! Set to zero so next MAXLOC doesn't 
                         ! find this one

        ! Now find nearest associated pixels and add to pixel list

        ! Look in 8 directions eight check is recursive
        ! groups all associated pixels together
        
        currentPixel => currentSource%pixelList
        nullify(currentPixel%previous)
        call nearestNeighbourNew(image, nx, ny, xpos, ypos, minSigma, currentPixel)
              
        ! Next maximum datapoint

        maxpos = MAXLOC(image)     
        maxpos(1) = maxpos(1) - 1
        maxpos(2) = maxpos(2) - 1
        if(image(maxpos(1), maxpos(2)) .lt. nSigma) exit linkSources
        
        tmp = image(maxpos(1), maxpos(2))

        xpos = maxpos(1)
        ypos = maxpos(2)

        ! Now move onto next source
          
        nSourceList = nSourceList + 1
        allocate(currentSource%next,stat=allocErr)
        call CheckAllocation(allocErr, "pixelList (getLinkedSources)")

        currentSource%next%previous => currentSource
     
        currentSource => currentSource%next

    end do linkSources     

end subroutine getLinkedSources
   
!*************************************************
! subroutine nearestNeighbourNew
! Check nearest pixels without recursive call
!
!*************************************************
subroutine nearestNeighbourNew(image, nx, ny, xpos, ypos, nSigma, outPixel)

    integer (kind=int32), intent(in) :: nx,ny
    real (kind=single), intent(inout) :: image(0 : nx - 1, 0 : ny - 1)
    integer (kind=int32), intent(in) :: xpos, ypos
    real (kind=single), intent(in) :: nSigma
    type(srcpixel), pointer :: outPixel, nextPixel

    integer  (kind=int32), allocatable :: there(:, :)
    type(srcPixel), pointer :: checkPixel
    integer (kind=int32) :: allocErr
    integer (kind=int32) :: k, l
    integer (kind=int32) :: minXpos, minYpos, maxXpos, maxYpos

!  print *,"nearestNeighbourNew"

    allocate(there(0 : nx - 1, 0 : ny - 1), stat = allocErr)
    call CheckAllocation(allocErr, "there") 

    there = 0
    !********************
    ! Do middle line
    !********************

    nextPixel => outPixel
    nextPixel%x = xpos
    nextPixel%y = ypos
    nullify(nextPixel%next)
    checkPixel => nextPixel

    loop: do

        call getMinMaxPosition(checkPixel%x, checkPixel%y, nx, ny, &
                               minXpos, maxXpos, minYpos, maxYpos)
     
        do k = minYpos, maxYpos
            do l = minXpos, maxXpos
                if(k .ne. checkPixel%x .and. l .ne. checkPixel%y) then
                    if(there(l, k) .eq. 0 .and. image(l, k) .gt. nSigma) then
                        there(l, k) = 1
                 
                        allocate(nextPixel%next, stat = allocErr)
                        call CheckAllocation(allocErr, "nextPixel%next (nearestNeighbourNew)")
                        nextPixel%next%previous => nextPixel
                        nextPixel => nextPixel%next
                        nextPixel%x = l
                        nextPixel%y = k
                        nullify(nextPixel%next)
                    end if
                end if
            end do
        end do

        !********************************************
        ! Move check position - if come to end exit
        !********************************************
        if(associated(checkPixel%next)) then
            checkPixel => checkPixel%next
        else
            exit loop
        end if

    end do loop

    !*******************************
    ! Removed used pixels from image
    !*******************************
    where(there .eq. 1) image = 0
  

    if(allocated(there)) then
        deallocate(there, stat = allocErr)
        call CheckDeallocation(allocErr, "there")
    end if
end subroutine nearestNeighbourNew

!*********************************************
! subroutine nearestNeighbour
! Check 8 sides for nearest pixels
!*********************************************
recursive subroutine nearestNeighbour(image, nx, ny, xpos, ypos, nSigma, &
     currentSource, currentPixel)

  integer (kind=int32), intent(in) :: nx,ny
  real (kind=single), intent(inout) :: image(0 : nx - 1, 0 : ny - 1)
  integer (kind=int32), intent(in) :: xpos, ypos
  real (kind=single), intent(in) :: nSigma
  type(src), pointer :: currentSource
  type(srcpixel), pointer :: currentPixel, nextPixel

  integer (kind=int32) :: minXpos, maxXpos, minYpos, maxYpos
  integer (kind=int32) :: allocErr

  integer (kind=int32) :: i, j

!  print *,"NearestBNeighbour"

  ! Make sure that we don't violate arrays
  ! This should be faster the MIN/MAX?

  call getMinMaxPosition(xpos, ypos, nx, ny, minXpos, maxXpos, minYpos, maxYpos)

  ! Find next significant pixel

  do i = minYpos, maxYpos
     do j = minXpos, maxXpos
        if(i .ne. xpos .and. j .ne. ypos) then
           if(image(j, i) .ge. nSigma) then

              ! This is asociated with source

              currentSource%numPixels = currentSource%numPixels + 1
              allocate(currentPixel%next,stat=allocErr)
              call CheckAllocation(allocErr, "pixel%next (nearestNeighbour)")
              
              nextPixel => currentPixel%next
              nextPixel%x = j
              nextPixel%y = i
              image(j, i) = 0.  ! reset image so ignore this pixel
              nullify(nextPixel%next)

              ! Now go down a level (recursive)

              currentPixel => nextPixel
              call nearestNeighbour(image, nx, ny, j, i, nSigma, currentSource, &
                   currentPixel)

           end if
        end if
     end do
  end do

end subroutine nearestNeighbour

!*************************************************************
! subroutine nearestNeighbourImage
! Check 8 sides for nearest pixels
! In this case add data to level image matching source number
!*************************************************************
recursive subroutine nearestNeighbourImage(image, nx, ny, levelImage, &
     minXpos, maxXpos, minYpos, maxYpos, xpos, ypos, nSigma, sourceNumber)
 
  integer (kind=int32), intent(in) :: nx, ny
  real (kind=single), intent(in) :: image(0 : nx - 1, 0 : ny - 1)
  integer (kind=int16), intent(inout) :: levelImage(0 : nx - 1, 0 : ny - 1)
  integer (kind=int32), intent(in) :: minXpos, maxXpos, minYpos, maxYpos
  integer (kind=int32), intent(in) :: xpos, ypos
  real  (kind=single), intent(in) :: nSigma
  integer (kind=int32), intent(in) :: sourceNumber

  integer (kind=int32) :: minX, maxX, minY, maxY
  integer (kind=int32) :: i, j

!  print *,"NearestNeighbourImage"

  ! Ensure we havn't strayed from allowed bounds

  if(xpos .lt. minXpos .or. xpos .gt. maxXpos) return
  if(ypos .lt. minYpos .or. ypos .gt. maxYpos) return

  ! Make sure that we don't violate arrays
  ! This should be faster the MIN/MAX?

  call getMinMaxPosition(xpos, ypos, nx, ny, minX, maxX, minY, maxY)

  ! Find next significant pixel

  do i = minYpos, maxYpos
     do j = minXpos, maxXpos
        if(i .ne. xpos .and. j .ne. ypos) then
           if(image(j, i) .ge. nSigma .and. levelImage(j, i) .eq. 0) then

              levelImage(j, i) = sourceNumber
              
              ! This is asociated with source

              call nearestNeighbourImage(image, nx, ny, levelImage, minXpos, &
                   maxXpos, minYpos, maxYpos, j, i, nSigma, sourceNumber)

           end if
        end if
     end do
  end do

end subroutine nearestNeighbourImage
!***********************************************
! subroutine getMinMaxPosition
!***********************************************
subroutine getMinMaxPosition(xpos,ypos,nx,ny,minXpos,maxXpos,minYpos,maxYpos)

    integer (kind=int32), intent(in) :: xpos, ypos
    integer (kind=int32), intent(in) :: nx, ny
    integer (kind=int32), intent(out) :: minXpos, maxXpos
    integer (kind=int32), intent(out) :: minYpos, maxYpos

!  print *,"getMinMaxPosition"

    if(xpos .gt. 0 .and. xpos .lt. nx - 1) then
        minXpos = xpos-1
        maxXpos = xpos+1
    else if(xpos .eq. 0) then
        minXpos = 0
        maxXpos = 1
    else
        minXpos = nx-2
        maxXpos = nx-1
    end if

    if(ypos .gt. 0 .and. ypos .lt. ny - 1) then
        minYpos = ypos - 1
        maxYpos = ypos + 1
    else if (ypos .eq. 0) then
        minYpos = 0
        maxYpos = 1
    else
        minYpos = ny - 2
        maxYpos = ny - 1
    end if

end subroutine getMinMaxPosition

!********************************************************
! subroutine smoothIMage
! Routine to smooth image for background determination
!********************************************************
subroutine smoothImage(image, nx, ny, background, smoothSize)
 
    integer (kind=int32), intent(in) :: nx, ny
    real (kind=single), intent(in)    :: image(0 : nx - 1, 0 : ny - 1)
    real (kind=single), intent(inout) :: background(0 : nx - 1, 0 : ny - 1)
    real (kind=single), intent(in)    :: smoothSize

! internal variables

    integer (kind=int32)            :: height
    integer (kind=int32)            :: i

! For top hat
  
    real (kind=single), allocatable   :: scan(:)
    integer (kind=int32), allocatable :: nscan(:)

    integer (kind=int32)            :: allocErr

!  print *,"SmoothImage"

! for gaussian smooth 

    height = smoothSize
     
  ! Get scan array
  
    allocate(scan(-height : height), stat = allocErr)
    call CheckAllocation(allocErr, "scan")

    allocate(nscan(-height : height), stat = allocErr)
    call CheckAllocation(allocErr, "nscan")

  ! smoothTopHat deals with one X position at a time

    do i = 0, nx - 1
        call MovingBox(image, nx, ny, background, height, height, i, scan, nscan, 1)
    end do
  
  ! deallocate memory
    if(allocated(scan)) then
        deallocate(scan, stat = allocErr)
        call CheckDeallocation(allocErr, "scan")
    end if

    if(allocated(nscan)) then
        deallocate(nscan, stat = allocErr)
        call CheckDeallocation(allocErr, "nscan")
    end if

end subroutine smoothImage

!******************************************************************************
! subroutine MovingBox
! Top hat smoothing routine - does a strip in Y direction (best array access)
!******************************************************************************
subroutine MovingBox(image, nx, ny, backgroundTot, width, height, startX, scan, &
                     nscan,type)

    integer (kind=int32), intent(in) :: nx, ny
    real (kind=single), intent(in)    :: image(0 : nx - 1, 0 : ny - 1)
    real (kind=single), intent(inout) :: backgroundTot(0 : nx - 1, 0 : ny - 1)

    integer (kind=int32), intent(in) :: width, height     ! Size of smoothing area
    integer (kind=int32), intent(in) :: startX           ! starting point
    real (kind=single), intent(inout) :: scan(-height : height)
    integer (kind=int32), intent(inout) :: nscan(-height : height)
    integer (kind=int32), intent(in) :: type
    !***********************
    ! internal variables
    ! for top hat smooth
    !***********************
    integer (kind=int32)            :: i, j ,k
    real(kind=single)                :: running_total
    integer (kind=int32)             :: nrunning_total
    integer (kind=int32)            :: minXpos, maxXpos
    integer (kind=int32)            :: minYpos, maxYpos
    integer (kind=int32)            :: clearStart, clearEnd
    real (kind=single)                :: scanval

!  print *,"MovingBox"

    !************************************
    ! initialise arrays for this strip
    !************************************
    scan = 0
    nscan = 0

    ! X min and max's are fixed
     
    minXpos = MAX(startX - width, 0)
    maxXpos = MIN(startX + width, nx -1)

 
    !**************************************************************
    ! Get start and end of clear (ie. don't need MIN/MAX checks )
    !**************************************************************
    clearStart = MIN(height + 1, ny)

    if(height .gt. ny) then
        clearEnd = -1
    else
        clearEnd = ny - height - 1
    end if

    !******************************************
    ! At start need to build up running total  
    ! First one
    !******************************************
    minYpos = 0
    maxYpos = MIN(height, ny - 1)


    running_total= 0.0
    nrunning_total= 0


    do j = minYpos, maxYpos     
        scanval = 0.0
        forall (k = minXpos : maxXpos) scanval = scanval + image(k, j)
        
        scan(j) = scanval
        nscan(j) = maxXpos - minXpos + 1
        running_total = running_total + scanval
        nrunning_total = nrunning_total + nscan(j)

    end do


  
    if(type .eq. 1) then
        backgroundTot(startX, 0) = running_total / nrunning_total
    else
        backgroundTot(startX, 0) = running_total
    end if

  
    do i = 1, clearStart - 1
     
        ! Shift down arrays
     
        do j = minYpos,  maxYpos
            scan(j - 1) = scan(j)
            nscan(j - 1) = nscan(j)
        end do
     
        ! reset min/max values
     
        minYpos = MAX(i - height, 0)
        maxYpos = MIN(i + height, ny - 1)

        !***********************
        ! Add in new entry
        !***********************     
        scanval = 0.0
        forall (k = minXpos : maxXpos) scanval = scanval + image(k, maxYpos)
        
        minYpos = minYpos - i
        maxYpos = maxYpos - i
        scan(maxYpos) = scanval
        nscan(maxYpos) = maxXpos - minXpos + 1
        running_total = running_total + scanval
        nrunning_total = nrunning_total + nscan(maxYpos)
 
        if(type .eq. 1) then
            backgroundTot(startX, i) = running_total / nrunning_total
        else
            backgroundTot(startX, i) = running_total
        end if
    
    end do
 

    !********************************
    ! Now do the rest (not MIN/MAX)
    !********************************
    if(clearEnd .eq. -1) return
  

    ! Make sure we can hold whole smoothing area within data

    if(clearStart .lt. clearEnd) then
        do i = clearStart, clearEnd
     
            ! Shift down arrays and remove bottom line

            running_total = running_total - scan(-height)
            nrunning_total = nrunning_total - nscan(-height)
            do j = -height + 1, height
                scan(j - 1) = scan(j)
                nscan(j - 1) = nscan(j)
            end do
            !*******************
            ! Add in new entry
            !******************
        
            maxYpos = i + height
            scanval = 0.0
            forall (k = minXpos : maxXpos) scanval = scanval + image(k, maxYpos)
            
            scan(height) = scanval
            nscan(height) = maxXpos - minXpos + 1
            running_total = running_total + scanval
            nrunning_total = nrunning_total + nscan(height)
       
        
            if(type .eq. 1) then
                backgroundTot(startX, i) = running_total / nrunning_total
            else
                backgroundTot(startX, i) = running_total
            end if

        end do
    end if

    !**********************
    ! And now do end part
    !**********************
  
    do i = clearEnd + 1, ny - 1
        running_total = running_total - scan(height - i + clearEnd + 1)
        nrunning_total = nrunning_total - nscan(height - i + clearEnd + 1)

        if(type .eq. 1) then
            backgroundTot(startX, i) = running_total / nrunning_total
        else
            backgroundTot(startX, i) = running_total
        end if
    end do

end subroutine MovingBox

!**************************************************************
! function computeMedian
! computes the median of an image over the specific limits
!**************************************************************
function computeMedian(image, nx, ny, i1, i2, j1, j2) result (median)

  
    integer (kind=int32) :: nx, ny
    real (kind=single):: image(0 : nx - 1, 0 : ny - 1)
    integer (kind=int32) :: i1, i2, j1, j2
    integer (kind=int32) :: i, j
    real (kind=single), allocatable :: tempArray(:)
    integer (kind=int32) :: npts
    real (kind=single):: median

    integer (kind=int32)            :: allocErr


    npts = (i2 - i1 + 1) * (j2 - j1 + 1)
    allocate(tempArray(npts), stat = allocErr)
    call CheckAllocation(allocErr, "tempArray")

    tempArray = 0
    npts = 0

    do i = i1, i2
        do j = j1, j2
            if(image(i, j) .ne. 0.0) then
                npts = npts + 1
                tempArray(npts) = image(i, j)
            end if
        end do
    end do

    median = MedianValue(tempArray, npts)

    deallocate(tempArray, stat = allocErr)
    call CheckDeallocation(allocErr, "tempArray")
  
end function computeMedian

!*******************************
! computeMode
!******************************
subroutine computeMode(image, nx, ny, nSigma, mode)

    integer (kind=int32), intent(in) :: nx, ny
    real (kind=single), intent(inout)    :: image(0 : nx - 1, 0 : ny - 1)
    real (kind=single), intent(in)    :: nSigma
    real (kind=single), intent(out)   :: mode

    integer (kind=int32) :: i, j, k

    real (kind=single), allocatable :: list(:), oldList(:)
    integer (kind=int32) :: nlist
    real (kind=single) maxImage
    integer (kind=int32) :: allocStat
    integer (kind=int32) :: m

    real (kind=single):: sigma
    real (kind=single):: mean, median
    real (kind=single):: variance, originalSigma, testSigma
    real (kind=single):: lowValue, highValue
    real (kind=single):: imageVal
    integer (kind=int32) :: intImageVal
    integer (kind=int32) :: nhist
    integer (kind=int32), allocatable :: histogram(:)
    integer (kind=int32) :: maxHistPos(1)
    logical (kind=int32) :: allocHist
    integer (kind=int32) :: allocErr
    integer (kind=int32) :: iMax

    allocate(list(nx * ny), stat = allocStat)
    call CheckAllocation(allocStat, "list")


    allocate(oldList(nx * ny), stat = allocStat)
    call CheckAllocation(allocStat, "oldlist")

!---VNY 24.10.2001 A protection against negative values of nhist is introduced
!    nhist = MAXVAL(image)
    maxImage = MAXVAL(image)
    nhist = int(maxImage)

    allocHist = .false.
!--- VNY 25.10.2001 Additional protection agains too big values of nhist
!    if(nhist .gt. 30) then
    if(nhist .gt. 30 .and. nhist .lt. 5000) then
        allocate(histogram(0 : nhist + 1), stat = allocStat)
        call CheckAllocation(allocStat, "histogram")
        histogram = 0
        allocHist = .true.
!---VNY 24.10.2001
     else
        call warning("ComputeMode: historgam allocation problem")
        write(errstr, *) "nhist=", nhist
        call message(AppMsg, Verbose)
    end if
 
    nlist = 0
    mean = 0.0

    do i = 0, ny - 1
        do j = 0, nx - 1

            nlist = nlist + 1
            imageVal = image(j, i)
            list(nlist) = imageVal 
            mean = mean + imageVal

            if(allocHist) then
                intImageVal = int(imageVal)

!---VNY 04.07.2001 Added protection against subscript of the range
                if ((intImageVal .lt. nhist) .and. (intImageVal .ge. 0) ) then 
                  histogram(intImageVal) = histogram(intImageVal) + 1
               end if
            end if
        end do
    end do 
 


    mean = mean / nlist


    if(allocHist) then
        maxHistPos = MAXLOC(histogram)

        mode = maxHistPos(1)
    else
       mode = mean ! Correction VNY 03.08.2001
    end if  


 
    !**************************************
    ! Get variance and std deviation
    !**************************************


    variance = 0.0
    forall (k = 1 : nlist) variance = variance + (list(k) - mode) ** 2
    
    variance = variance / (nlist - 1)
 
    originalSigma = sqrt(variance)
    lowValue = mode - nSigma * originalSigma
    highValue = mode + nSIgma * originalSigma


    !**********************************************
    ! reject data at more than nSigma from data
    !**********************************************
    testSigma = originalSigma

    !*********************************************************
    !--- VNY 06.08.2001 limiting the number of sigmaClip loops
    !*********************************************************
    iMax = 15000
    sigmaClip: do i = 1, iMax 

       

        oldlist = list
        m = 0
        mean = 0.0
        do k = 1, nlist
            if(oldList(k) .ge. lowValue .and. oldList(k) .le. highValue) then
                m = m + 1
                list(m) = oldList(k)
                mean = mean + list(m)
            end if
        end do     
 

        nlist = m
        if(nlist .lt. 20) exit sigmaClip
       
        mean = mean / nlist
        nhist = MAXVAL(list)

        if(nhist .gt. 30 .and. allocHist) then
            histogram = 0
            forall  (k = 1 : nlist)
                intImageVal = int(list(k))

                histogram(intImageVal) = histogram(intImageVal) + 1
            end forall
            maxHistPos = MAXLOC(histogram)
            mode = maxHistPos(1)        
        else
            mode = mean
        end if
 
        !*******************************
        ! Now do sigma clipping
        ! Get variance and std deviation
        !********************************
        variance = 0.0
        forall (k = 1 : nlist) variance = variance + (list(k) - mode) ** 2
        
        variance = variance / (nlist - 1)
 
        sigma = sqrt(variance)
        lowValue = mode - nSigma * sigma
        highValue = mode + nSigma * sigma

        if(abs((sigma - testSigma) / sigma) .lt. 0.2) exit sigmaClip
        
        testSigma = sigma

    end do sigmaClip


!************************************************************************** 
!   VNY 06.08.2001 This code underestimates the background median value,
!   hence it is removed 
!**************************************************************************
    if(abs((sigma - originalSigma) / sigma) .gt. 0.2) then
        median = MedianValue(list, nlist)
        mode = 2.5 * median - 1.5 * mean
    else
        mode = mean
    end if
!*************************************************************************** 

    if(allocHist) then
        deallocate(histogram, stat = allocErr)
        call CheckDeallocation(allocErr, "histogram")
    end if

    deallocate(list, stat = allocErr)
    call CheckDeallocation(allocErr, "list")
  
    deallocate(oldList, stat = allocErr)
    call CheckDeallocation(allocErr, "oldList")
   

end subroutine computeMode
!***************************************************
! computeBackground
! Computes the background using a median box filter
!***************************************************
subroutine computeBackground(image, nx, ny, background, smBackground, &
     medianBackground, nSigma)

    integer (kind=int32), intent(in) :: nx, ny
    real (kind=single), intent(in)    :: image(0 : nx - 1, 0 : ny - 1)
    real (kind=single), intent(inout) :: background(0 : nx - 1, 0 : ny -1)
    real (kind=single), intent(in)    :: smBackground(0 : nx - 1, 0 : ny - 1)
    real (kind=single), intent(out)   :: medianBackground
    real (kind=single), intent(in)    :: nSigma
    real (kind=single)                :: sigmaBackground
    integer (kind=int8), parameter  :: nBox = 5
    real (kind=single), allocatable   :: tmpBackground(:, :)
    logical, allocatable :: mask(:, :)
    real (kind=single)                 :: rmsval, newrmsval
    integer (kind=int32)            :: allocErr

    allocate(tmpBackground(0 : nx - 1, 0 : ny - 1), stat = allocErr)
    call CheckAllocation(allocErr, "tmpBackground")

    allocate(mask(0 : nx - 1, 0 : ny - 1), stat = allocErr)
    call CheckAllocation(allocErr, "mask")

    tmpBackground = smBackground
    sigmaBackground = sqrt(medianBackground)

    !**********************************************************************
    ! Get rid of bright sources on scale of smoothing - use sigma clipping
    !
    ! sigma clip until got down to a reasoanble value
    !*********************************************************************
    rmsval = sqrt(SUM((smBackground - medianBackground) ** 2) / size(smBackground))

    sigma_clip: do 
        
        where ( (smBackground - medianBackground) .lt. nSigma * rmsval) 
            mask = .true.
        elsewhere
            mask = .false.
        end where

        newrmsval = rmsval
        if(count(mask) .gt. 1) then
            rmsval = sqrt(SUM((smBackground - medianBackground) ** 2, mask) / &
                         count(mask))

        else
            exit sigma_clip
        end if

        if((newrmsval - rmsval) / newrmsval .lt. 0.5) exit sigma_clip
        
    end do sigma_clip

    !****************************************
    ! Now put median value where sources are
    !****************************************
    where ( (smBackground - medianBackground) .gt. nSigma * rmsval) tmpBackground = medianBackground
    
    !*********************************************************
    ! Get rid of bright sources smaller than smoothing scale
    !*********************************************************
    where ( ((image - medianBackground) / sigmaBackground) .gt. nSigma) tmpBackground = medianBackground
    
    background = tmpBackground

    if(allocated(tmpBackground)) then
        deallocate(tmpBackground, stat = allocErr)
        call CheckDeallocation(allocErr, "tmpBackground")
    end if

    if(allocated(mask)) then
        deallocate(mask, stat = allocErr)
        call CheckDeallocation(allocErr, "mask")
    end if

end subroutine computeBackground
!*********************************************
!***************************************************
! AddReal32Cololumn
! Adds a column of real32 data to a fits table.
!***************************************************
subroutine AddReal32Column(tableSet, heading, comment, &
        size, array, nullValue)
    type(TableT), intent(in) :: tableSet
    character(len=*), intent(in) :: heading, comment
    integer, intent(in) :: size
    real (kind=single), intent(in), dimension(:) :: array
    real (kind=single), intent(in) :: nullValue
    real (kind=single), pointer :: real32Col(:)
    type(ColumnT) :: pColumn
    integer (kind=int32)i
    integer (kind=int32):: ndims(1) = (/1/)
 
    pcolumn = addColumn(tableSet, heading, real32, comment, ndims)
    real32Col => real32Data(pcolumn)

    do i = 1, size
        if (array(i) .eq. nullValue) then
            call setToNull(pcolumn, i - 1)
        else
            real32Col(i - 1) = array(i)
        end if
  
    end do
    call release(pcolumn)

end subroutine AddReal32Column
!***************************************************
! AddReal64Column
! Adds a column of real64 data to a fits  table.
!***************************************************
subroutine AddReal64Column(tableSet, heading, comment, &
        size, array, nullValue)
    type(TableT), intent(in) :: tableSet
    character(len=*), intent(in) :: heading, comment
    integer (kind=int32), intent(in) :: size

    real (kind=double), intent(in), dimension(1 : size) :: array
    real (kind=single), intent(in) :: nullValue
    real (kind=double), pointer :: real64Col(:)
    type(ColumnT) :: pColumn
    integer (kind=int32) :: i
    integer (kind=int32):: ndims(1) = (/1/)

    pcolumn = addColumn(tableSet, heading, real64, comment, ndims)
    real64Col => real64Data(pcolumn)

    do i = 1, size
        if (array(i) .eq. nullValue) then
            call setToNull(pcolumn, i - 1)
        else
            real64Col(i - 1) = array(i)
        end if
  
    end do
    call release(pcolumn)

end subroutine AddReal64Column
!***************************************************
! AddInt8Column
! Adds a column of integer8 data to a fits table.
!***************************************************
subroutine AddInt8Column(tableSet, heading, comment, &
        size, array)
    type(TableT), intent(in) :: tableSet
    character(len=*), intent(in) :: heading, comment
    integer (kind=int32), intent(in) :: size
    integer (kind=int8), intent(in), dimension(:) :: array
    integer (kind=int8), pointer :: int8Col(:)
    type(ColumnT) :: pColumn
    integer (kind=int32) :: i
    integer (kind=int32):: ndims(1) = (/1/)

    pcolumn = addColumn(tableSet, heading, integer8, comment, ndims)
    int8Col => int8Data(pcolumn)

    forall (i = 1 : size) int8Col(i - 1) = array(i)
    

    call release(pcolumn)

end subroutine AddInt8Column
!******************************************
! Subroutine SaveLevelImage
! Creates a file to store the image
!******************************************
subroutine SaveLevelImage(levelImage, nx, ny, fileName, heading)
    integer, intent(in) :: nx, ny    
 
    integer(kind=int16), intent(in), dimension(0 : nx - 1, 0 : ny - 1) :: levelImage
    character(len=*), intent(in) :: fileName
    character(len=*), intent(in) :: heading
    type(DataSetT)       :: levelImageSet 
    integer(kind=int16), dimension(:, :), pointer :: levelImageP
    type(ArrayT)         :: levelImageArray         ! array for background
    
    
    levelImageSet = DataSet(fileName, Create)

    levelImageArray = addArray(levelImageSet, heading, Integer16, &
                 shape(levelImage))

    levelImageP => int16Array2Data(levelImageArray)

    levelImageP = levelImage

    call release(levelImageArray)

    call release(levelImageSet)


end subroutine SaveLevelImage
!****************************************************
! Subroutine SaveBackgroundImage
!****************************************************
subroutine SaveBackgroundImage(backgroundImage, nx, ny, fileName, heading)
    integer (kind=int32), intent(in) :: nx, ny    
 
    real (kind=single), intent(in), dimension(0 : nx - 1, 0 : ny - 1) :: backgroundImage
    character(len=*), intent(in) :: fileName
    character(len=*), intent(in) :: heading
    type(DataSetT)       :: backGroundImageSet 
    real (kind=single), dimension(:, :), pointer :: backgroundImageP
    type(ArrayT)         :: backgroundImageArray         ! array for background
   
 
    backgroundImageSet = DataSet(fileName, Create)

    backgroundImageArray = addArray(backgroundImageSet, heading, real32, &
                 shape(backgroundImage))

    backgroundImageP => real32Array2Data(backgroundImageArray)

    backgroundImageP = backgroundImage

    call release(backgroundImageArray)

    call release(backgroundImageSet)

end subroutine SaveBackgroundImage
!*****************************************************************
! Subroutine FastChainChecks
! Vladimir's checks for fast-chain data
!****************************************************************
subroutine FastChainChecks(minX, maxX, minY, maxY, &
          nx, ny, image, sourcePars)
    integer (kind=int32), intent(in) :: minX, maxX, minY, maxY
    integer (kind=int32), intent(in) :: nx, ny
    real (kind=single), intent(in), dimension(0 : nx - 1, 0 : ny - 1) :: image
    type(sourceParams), intent(inout) :: sourcePars  

    integer (kind=int32) :: mnX, mxX, mnY, mxY, jmax, mmax
    real (kind=double) :: sumax, borderX, borderY
    real (kind=double), allocatable :: meanX( : )
    real (kind=double), allocatable :: meanY( : )
    integer (kind=int32) :: j, m
    integer (kind=int32) :: allocErr

    mnY=minY-1
    mxY=maxY-1
    mnX=minX-1
    mxX=maxX-1


    allocate (meanY(0:ny-1), stat=allocErr)
    call CheckAllocation(allocErr, "meanY")

    allocate(meanX(0:nx-1), stat=allocErr)
    call CheckAllocation(allocErr, "meanX")

    if (mxY .eq. ny-1 .or. mnY .eq. 0) then

        write(errstr,'(a)') "Correcting Y-position (proximity to the FOV Y-edge)"
        call message(AppMsg, Verbose)

                  
!---     VNY 03.09.2001 another correcting algorithm. Just looking for maximum
!        When the source is near the edge of the FOW, the moments could give 
!        a mistaken source position
        ! Looking for the maximum along Y
        ! Averaging along X
        meanX=0.0
        sumax=0.0
        jmax=mnX+1
        do j=mnX+1, mxX-1
            do m=mnY+1, mxY-1
                meanX(j)=meanX(j)+image(j,m)
            end do
            if (meanX(j) .gt. sumax) then
                sumax=meanX(j)
                jmax=j
            end if
        end do

                
        ! Looking for the maximum Y
        sumax=0.0
        mmax=mnY+1
        do m=mnY+1, mxY-1
            if (image(jmax,m) .gt. sumax) then 
                sumax=image(jmax, m)
                mmax=m
            end if
        end do
                 
        ! Calculate the corrected Y-position
        borderY=real(mmax)
        if (mmax .lt. ny-1 .and. mmax .gt. 0) then
                    
            borderY=borderY+0.5*(image(jmax,mmax+1)-image(jmax, mmax-1))/(image(jmax,mmax+1)+image(jmax,mmax-1))
        end if

!---VNY 20.09.2001 the sourcePars array corresponds to the number of subsources
!                      for the current source "i" the coordinate is changed for the first subsource                  
                  sourcePars%y=borderY+1.0
    end if


!--------VNY: the same for the X-coordinate 

    if (mxX .eq. nx-1 .or. mnX .eq. 0) then

        write(errstr,'(a)') "Correcting X-position (proximity to the FOV X-edge)"
        call message(AppMsg, Verbose) 
                  
                 
        !Looking for the maximum along X 
                  ! Averaging along Y
        meanY=0.0
        sumax=0.0
        mmax=mnY+1
        do m=mnY+1, mxY-1
            do j=mnX+1, mxX-1
                meanY(m)=meanY(m)+image(j,m)
            end do
            if (meanY(m) .gt. sumax) then
                sumax=meanY(m)
                mmax=m
            end if
        end do

 
        ! Looking for the maximum X
        sumax=0.0
        jmax=mnX+1
        do j=mnX+1, mxX-1
            if (image(j, mmax) .gt. sumax) then 
                sumax=image(j, mmax)
                jmax=j
            end if
        end do
 
        ! Calculate the corrected X-position
        borderX=real(jmax)
        if (jmax .lt. nx-1 .and. jmax .gt. 0) then
            borderX=borderX+0.5*(image(jmax+1,mmax)-image(jmax-1,mmax))/(image(jmax+1,mmax)+image(jmax-1,mmax))
        end if
!---VNY 20.09.2001 correcting a subsource index
        sourcePars%x=borderX+1.0
    end if

    deallocate (meanX)
    deallocate (meanY)

end subroutine FastChainChecks

end module omdetect_mod









