!******************************************************************
! Module omdetectnumerics_mod
!
! Contains various subroutines for numerical procedures.
!******************************************************************
module omdetectnumerics_mod
  use errorHandling
  use omdetectconstants_mod
  use sort_mod
contains
!********************************************************************
! Subroutine fitGaussian
! Fitting routine to get source parameters to assign pixel parameters
!********************************************************************
subroutine fitGaussian(image, nx, ny, levelImage, i1, i2, j1, j2, params)

    integer (kind=int32), intent(in) :: nx, ny
    real (kind=single), intent(in) :: image(nx, ny)
    integer  (kind=int32), intent(in) :: i1, i2, j1, j2
    integer  (kind=int32), intent(in) :: levelImage(i1 : i2, j1 : j2)
    real (kind=single), intent(out) :: params(6)
    real (kind=single):: fwhm_x, fwhm_y
    real (kind=single):: pmat(7, 6)
    real (kind=single):: ptest(6)
    real (kind=single):: yvec(7)
    real (kind=single):: frac(6)
    integer  (kind=int32) :: m,n
    integer  (kind=int32) :: maxvalue(2)
    integer  (kind=int32) :: iter
    logical(kind=bool) :: ok

    !**************************
    ! Set up frac array
    !**************************
    frac(1) = 1.05       ! fg
    frac(2) = 1.05       ! sig_x
    frac(3) = 1.05       ! sig_y
    frac(4) = 1.01       !pa
    frac(5) = 1.0005     ! x
    frac(6) = 1.0005     ! y
  
  ! set up initial guesses for gaussian

    fwhm_x = real(i2 - i1) / 1.5
    fwhm_y = real(j2 - j1) / 1.5
    fwhm_x = max(fwhm_x, 0.5)
    fwhm_y = max(fwhm_y, 0.5)
  
    params(1) = SUM(image(i1 : i2, j1 : j2))
    params(2) = fwhm_x / 2.35
    params(3) = fwhm_y / 2.35
    params(4) = 0.0
    maxvalue = MAXLOC(image(i1 : i2, j1 : j2))
    params(5) = real(maxvalue(1)) + i1
    params(6) = real(maxvalue(2)) + j1

 ! The following code defines the nodes of the simplex
  
    pmat(1, 1) = params(1)                        
    pmat(1, 2) = params(2)
    pmat(1, 3) = params(3)
    pmat(1, 4) = params(4)
    pmat(1, 5) = params(5)
    pmat(1, 6) = params(6)
  
    ! define nodes of simplex
  
    do m = 2 , 7
        forall (n = 1: 6) pmat(m, n) = params(n)
        pmat(m, m - 1) = pmat(m, m - 1) * frac(m - 1)
    end do
  
    do m = 1 , 7
        forall (n = 1 : 6) ptest(n) = pmat(m, n)
        yvec(m) = gauss2d(ptest, image, nx, ny, levelImage, i1, i2, j1, j2)
    end do


    !*****************************************************
    ! Fit the gaussian function, gauss2d, to the image 
    ! using the downhill simplex method
    !****************************************************
    call simplex(pmat, yvec, 7, 6, 6, 1.e-4, gauss2d, iter, image, nx, ny, &
               levelImage, i1, i2, j1, j2, ok)
  
    ! The solution is taken as the mean of the simplex nodes
!write(*, *) "Input source parameters are"
!do m = 1, 6
!    write(*, *) m, ") ", params(m)
!end do

  
    params = 0.0
    do m = 1, 7
        forall (n = 1 : 6) params(n) = params(n) + pmat(m, n)
    end do
    params = params  / 7.0

end subroutine fitGaussian
!***************************************************************
! simplex
! downhill simplex routine
!***************************************************************
subroutine simplex(pmatrix, yvec, msize, nsize, ndimen, simptol, exfunk, iter,&
                   osw, nx, ny, level, i1, i2, j1, j2 ,ok)

    logical (kind=bool) :: ok

    integer (kind=int32), parameter :: nmax = 20
    integer (kind=int32), parameter :: itmax = 1000
    real (kind=single), parameter    :: beta = 0.5
    real (kind=single), parameter    :: alpha = 1.0
    real (kind=single), parameter    :: gamma = 2.0
  
    integer (kind=int32) :: i1, i2, j1, j2
    integer (kind=int32) :: level(i1 : i2, j1 : j2)

    integer (kind=int32) :: msize
    integer (kind=int32) :: nsize
    integer (kind=int32) :: mpts
    integer (kind=int32) :: ndimen
    integer (kind=int32) :: iter
    integer (kind=int32) :: ilo
    integer (kind=int32) :: j

    integer (kind=int32) :: nx
    integer (kind=int32) :: ny
    real (kind=single)    :: osw(nx, ny)

    real (kind=single):: pmatrix(msize, nsize)
    real (kind=single):: yvec(msize)
    real (kind=single):: pr(nmax)
    real (kind=single):: prr(nmax)
    real (kind=single):: pbar(nmax)

    integer (kind=int32) :: ihi
    integer (kind=int32) :: inhi
    integer (kind=int32) :: i
    real (kind=single)   :: simptol
    real (kind=single)   :: rtol 
    real (kind=single)   :: ypr
    real (kind=single)   :: yprr

    real (kind=single)   :: exfunk
    external exfunk

    ok = .true.

    mpts = ndimen + 1
    iter = 0

    do while (.true.)
        ilo = 1

        if( yvec(1) .gt. yvec(2) ) then
            ihi = 1
            inhi = 2
        else
            ihi = 2
            inhi = 1
        end if

        do  i = 1 , mpts
            if (yvec(i) .lt. yvec(ilo)) ilo = i
            if (yvec(i) .gt. yvec(ihi)) then
                inhi = ihi
                ihi = i
            else if (yvec(i) .gt. yvec(inhi)) then
                if(i .ne. ihi) inhi = i
            end if
        end do

        rtol = 2.0 * abs(yvec(ihi) - yvec(ilo)) / (abs(yvec(ihi)) + abs(yvec(ilo)))

        if (rtol .lt. simptol) EXIT

        if (iter .eq. itmax) then
            ok = .false.
            EXIT
        end if
        iter = iter + 1

        pbar = 0.0

        do i = 1,  mpts 
            if (i .ne. ihi) then 
                forall (j = 1 :ndimen) pbar(j) = pbar(j) + pmatrix(i, j)
            end if 
        end do           
  

        pbar = pbar / real(ndimen)

        forall (j = 1 : ndimen) pr(j) = (1.0 + alpha) * pbar(j) - alpha * pmatrix(ihi, j)
     

        ypr = exfunk(pr, osw, nx, ny, level, i1 ,i2, j1 ,j2)

        if (ypr .le. yvec(ilo)) then

            forall  (j = 1 : ndimen) prr(j) = gamma *pr(j) + (1. - gamma) * pbar(j)
       

            yprr = exfunk(prr, osw, nx, ny, level, i1, i2, j1, j2)

            if (yprr .lt. yvec(ilo)) then
                forall (j = 1 : ndimen) pmatrix(ihi, j) = prr(j)
           
                yvec(ihi) = yprr
        else
           forall (j = 1 : ndimen) pmatrix(ihi, j) = pr(j)
           
           yvec(ihi) = ypr
        end if

     else if (ypr .ge. yvec(inhi)) then

        if (ypr .lt. yvec(ihi)) then
           forall (j = 1 : ndimen) pmatrix(ihi, j) = pr(j)
           
           yvec(ihi) =  ypr
        end if

        forall (j = 1 : ndimen) prr(j) = beta * pmatrix(ihi, j) + (1.0 - beta) * pbar(j)
       

        yprr = exfunk(prr, osw, nx, ny, level, i1, i2, j1, j2)

        if (yprr .lt. yvec(ihi)) then
           forall (j = 1 : ndimen) pmatrix(ihi, j) = prr(j)
           
           yvec(ihi) = yprr
        else
           do i = 1, mpts
              if(i .ne. ilo)then
                 forall (j = 1 : ndimen)
                    pr(j) = 0.5 * (pmatrix(i, j) + pmatrix(ilo, j))
                    pmatrix(i, j) = pr(j)
                 end forall
                 yvec(i) = exfunk(pr, osw, nx, ny, level, i1, i2, j1, j2)
              end if
           end do
        end if

     else

        forall (j = 1 : ndimen) pmatrix(ihi, j) = pr(j)
        

        yvec(ihi) = ypr

     end if

  end do

end subroutine simplex
!*******************************************************
! gauss2d
!*******************************************************
real(kind=single) function gauss2d(param, osw, nx, ny, level, i1, i2, j1, j2)

    implicit none

    real (kind=single) :: param(*)                  ! the Gaussians parameters
    integer (kind=int32) ::  nx, ny                  ! the osw image size
    real (kind=single) :: osw(nx, ny)                ! the osw data array
    integer (kind=int32) :: i, j, n                  ! counters
    real (kind=single) :: flux                      ! the source flux
    real (kind=single) :: x, y                       ! the x,y position of a pixel
    real (kind=single) :: x0, y0                     ! the x,y position of the gaussian
    real (kind=single) :: sig_x, sig_y              ! the sigma of the gaussians
    real (kind=single) :: pa
    real (kind=single) :: tot                       ! total
    real (kind=single) :: f                         ! flux
    real (kind=single) :: ex1, ex2, xDash, yDash, delta, r
    integer (kind=int32) :: i1, i2, j1, j2
    integer (kind=int32):: level(i1 : i2, j1 : j2)
    real (kind=single):: oswval, constant 

    ! set up the Gaussian parameters
    flux =  param(1)
    sig_x = param(2)
    sig_y = param(3)
    pa =   param(4)
    x0 =   param(5)
    y0 =   param(6)
    tot = 0.0
    n = 0
    

    constant = 0.159154943 * flux /   (sig_x * sig_y)

    if (sig_x * sig_y .ne. 0.0) then
        do i = j1 , j2
            y = real(i) + 0.5
           
            do j = i1 , i2
                if(level(j, i) .gt. 0) then
                    x = real(j) + 0.5
                    r = sqrt((x - x0) ** 2 + (y - y0) ** 2)
                    delta = atan2(y - y0, x - x0)
              
                    xDash = r * cos(delta - pa)
                    yDash = r * sin(delta - pa)
              
                    ex1 = -((xDash / sig_x) ** 2)
              
                    ex2 = -((yDash / sig_y) ** 2)
                           
              
                    f = constant * exp(0.5 * (ex1 + ex2))
              
                    oswval = osw(j, i)
                    if(oswval .gt. 0) then
                        tot = tot + (f - oswval) ** 2 / oswval
                    else if(oswval .eq. 0) then
                        tot = tot + f * f
                    end if
              
                    n = n + 1
                end if
            end do
        end do
    else
        tot = 1.0e20
    end if

    gauss2d  = tot
end function gauss2d

!********************************
! function Angle
! 
!********************************
real function Angle(x, y, xy)
    real (kind=single), intent(in) :: x, y, xy
    real (kind=single) :: theta1, theta2 

    if(x .ne. y) then
        theta1 = atan(2 * xy / (x - y)) / 2.0
        theta2 = theta1 + HALF_PI
        if(theta2 .gt. HALF_PI) theta2 = theta2 - PI
        if(xy .ge. 0.0) then
            if(theta1 .ge. 0.0) then
                Angle = theta1
            else
                Angle = theta2
            end if
        else
            if(theta1 .lt. 0.0) then
                Angle = theta1
            else
                Angle = theta2
            end if
        end if
    else
        Angle = sign(real(HALF_PI), xy) / 2.0
    end if
end function Angle

!***********************************************
! function gammln
!***********************************************
real(kind=double) function gammln(xx)

    implicit none
    real (kind=single), intent(in) :: xx
    real(kind=double) :: tmp, x, y, y0
    real(kind=double), save :: cof(6) = (/76.18009172947149_double, -86.50532032941677_double, &
                                          24.01409824083091d0, -1.231739572450155_double, &
                                          0.1208650973866179d-2, -0.5395239384953d-5/)  
    real(kind=double), save :: stp = 2.5066282746310005_double
    real(kind=double) :: ser = 1.000000000190015_double
    integer (kind=int32) :: j

    x = xx
    y0 = x
    tmp = x + 5.5_double
    tmp = (x + 0.5_double) * log(tmp) - tmp
  
 
    do j = 1, 6
        y = y0 + dble(j)
        ser = ser + cof(j) / y
    end do
    gammln = tmp + log(stp * ser / x)
  
end function gammln

!*********************************************
! function lnFactrl
! Log of n!
!*********************************************
real function lnFactrl(n) result(output)

    implicit none

    integer (kind=int32), intent(in)  :: n
  
    ! internal variables 
    real (kind=single), save :: a(100) = -1.0

  
    if(n .lt. 0)then
        call fatal("lnFactrl","Incorrect input n < 0")
    else if(n .le. 99) then
        if(a(n + 1) .lt. 0) a(n + 1) = gammln(n + 1.0)
        output = a(n + 1)      
    else
        output = gammln(n + 1.0)
    end if

end function lnFactrl

! errf - a=0.5 is explicit

!***************************************************************
! function newerrf
! New error function routine that copes with n sighma better
!
!***************************************************************
real(kind=double) function newerrf(x) result(resultval)

    real (kind=single), intent(in) :: x
    real (kind=double) :: double_x
    real (kind=double) :: z, t, expval
    
    !**************************************
    ! Use erfc function to get probability
    !**************************************
    double_x = x / sqrt(2.0_double)
    z = abs(double_x)
    t = 1.0_double / (1.0_double + 0.5_double * z)
    expval = -z * z - 1.26551223_double + t * (1.00002368_double + &
           t * (0.37409196_double + t * (0.09678418_double + &
           t * (-0.18628806_double + t * (0.27886807_double + &
           t * (-1.13520398_double + t * (1.48851587_double + &
           t * (-0.82215223_double + t * 0.17087277_double))))))))
    resultval = exp(log(t) + expval)

    if(resultval .lt. 0.0_double) resultval = 2.0_double - resultval

end function newerrf
!***************************************************
! function errfx
!
!***************************************************
real(kind=double) function errf(x) result(output)

    real (kind=single):: x
    real (kind=double) :: gammcf, gamser, xx

!  print *,"errf"

    xx = dble(x)
    if(xx .lt. 0.0_double) then
        call fatal("gammp","Incorrect arguments")
    end if
    if(xx .lt. 1.5_double) then
        call lnGserErf(gamser, xx * xx)
        output = exp(gamser)
    else
        call lnGcfErf(gammcf, xx * xx)
        output = 1 - exp(gammcf)
    end if

end function Errf
!***************************************
! subroutine lnGserErf 
! gser routine with a=0.5 explicit
!***************************************
subroutine lnGserErf(gamser, x) 

    real (kind=double) :: gamser,x

    integer (kind=int32) :: n
    real (kind=double) :: ap, del, sum
    real (kind=double) :: output
    real (kind=double) :: gln = 0.572365_double ! the value of gammln(0.5)


    if(x .le. 0.0_double) then
        if(x .lt. 0.0_double) then
            call fatal("gser_erf", "x < 0 in gser")
        end if
        output = 0.0_double
        return
    end if
  
    gamser = -1d8
    ap = 0.5_double
    sum = 1.0+double / ap
    del = sum
    erf_loop: do n = 1, 100
        ap =ap + 1.
        del = del * x / ap
        sum = sum + del
        if(abs(del) .lt. abs(sum) * 3.d-7) then
            gamser = log(sum) +(-x + 0.5_double *log(x) - gln)
            exit erf_loop
        end if
    end do erf_loop

    if(gamser .eq. -1d8) call fatal("gser_erf","Number of iteration exceeded")
  

end subroutine lnGserErf
!************************************************
! subroutine lnGcfErf
! gcf routine with a=0.5 explicit
!************************************************
subroutine lnGcfErf(gammcf, x)

    real (kind=double) :: gammcf, x
    integer (kind=int32) :: i
    real (kind=double) :: an, b, b0, c, d, del, h

!  gln = detectGammln(0.5)
!  make gln the value of gammln(0.5)

    real (kind=double) ::  gln = 0.572365_double

    gammcf = -1d8

    b0 = x + 0.5_double
    c = 1d30
    d = 1.0_double / b
    h = d

    gcf_loop: do i = 1, 100
        an = -i * (i - 0.5_double)
        b = b0 + dble(2 * i)
        d = an * d + b
        if(abs(d) .lt. 1d-30) d = 1d-30
     
        c = b + an / c
        if(abs(c) .lt. 1d-30) c = 1d-30
     
        d = 1.0_double / d
        del = d * c
        h = h * del
        if(abs(del-1.) .lt. 3.d-7) then
        gammcf = (-x + 0.5_double * log(x) - gln) + log(h)
     end if
  end do gcf_loop
  if(gammcf .eq. -1d8) then
     call fatal("gcf_erf","Number of interation exceeded")
  end if

end subroutine lnGcfErf
!*****************************************************************
! function getSigmaFromProb
! Routine to get sigma for a given (gaussian) probability
! This prob won;t work since errf can't get beyond 7 sigma or so...
!******************************************************************
real function getSigmaFromProb(prob, guess) result(output)

    real (kind=single), intent(in) :: prob, guess
    real (kind=single) :: lowRange, highRange

    !*******************************************************
    ! error function gives prob for given sigma so need to 
    !find correct sigma use failsafe method
    !*******************************************************
    lowRange = guess * 0.1
    highRange = guess * 10

    output = rtsafe(lowRange, highRange, 1e-4, prob)

end function getSigmaFromProb

!*******************************************
! subroutine getFunction
! Returns the function (root) to get sigma
!*******************************************
subroutine getFunction(x1, f, df, prob)

    real (kind=single), intent(in) :: x1, prob
    real (kind=single), intent(out) :: f, df
    real (kind=single):: inx


    f = Errf(x1)
    inx = x1 * 1.00001
    df = (Errf(inx) - f) / (0.00001 * x1)
    f = f + prob - 1

end subroutine getFunction
!***********************************
! function rtsafe
!***********************************
real function rtsafe(x1, x2, xacc, prob) result(output)

    real (kind=single), intent(in) :: x1, x2, xacc, prob
    integer (kind=int32) :: j
    real (kind=single):: df, dx, dxold, f, fh, fl, temp, xh, xl


    call getFunction(x1, fl, df, prob)
    call getFunction(x2, fh, df, prob)
  
    if((fl .gt. 0.0 .and. fh .gt. 0.0) .or. (fl .lt. 0.0 .and. fh .lt. 0.0)) then
        output = x2
        return
    else if(fh .eq. 0) then
        output = x2
        return
    else if(fl .lt. 0.0) then
        xl = x1
        xh = x2
    else
        xl = x2
        xh = x1
    end if

    output = 0.5 * (x1 + x2)
    dxold = abs(x2 - x1)
    dx = dxold
    call getFunction(output, f, df, prob)
    do j = 1, 100
        if(((output - xh) * df -f) * ((output - xl) * df - f) .ge. 0 .or. &
            abs(2.0 * f) .gt. abs(dxold * df)) then
            dxold =dx
            dx = 0.5 * (xh - xl)
            output = xl + dx
            if(xl .eq. output) return
        else
            dxold = dx
            dx = f / df
            temp = output
            output = output - dx
            if(temp .eq. output) return
        end if
        if(abs(dx) .lt. xacc) return
        call getFunction(output, f, df, prob)
        if(f .lt. 0.0) then
            xl = output
        else
            xh = output
        end if
    end do 

    call fatal("rtsafe","Too many iterations")

end function rtsafe



!***************************
! Function AllocateInt8
!***************************
function AllocateInt8(il, iu, name) result(arrayPtr)
    integer, intent(in) :: il, iu
    character(len=*), intent(in) :: name
    integer(kind=int8), dimension(:), pointer :: arrayPtr
    integer :: status
    
    allocate(arrayPtr(il : iu), stat = status)
    call CheckAllocation(status, name)
    arrayPtr = 0
end function AllocateInt8

!***************************
! Function AllocateInt32
!***************************
function AllocateInt32(il, iu, name) result(arrayPtr)
    integer, intent(in) :: il, iu
    character(len=*), intent(in) :: name
    integer(kind=int32), dimension(:), pointer :: arrayPtr
    integer :: status
    
    allocate(arrayPtr(il : iu), stat = status)
    call CheckAllocation(status, name)
    arrayPtr = 0
end function AllocateInt32

!***************************
! Function Allocate2DInt16
!***************************
function Allocate2DInt16(il, iu, jl, ju, name) result(arrayPtr)
    integer, intent(in) :: il, iu, jl, ju
    character(len=*), intent(in) :: name
    integer(kind=int16), dimension(:, :), pointer :: arrayPtr
    integer :: status
    
    allocate(arrayPtr(il : iu, jl : ju), stat = status)
    call CheckAllocation(status, name)
    arrayPtr = 0
end function Allocate2DInt16

!***************************
! Function Allocate2DInt32
!***************************
function Allocate2DInt32(il, iu, jl, ju, name) result(arrayPtr)
    integer, intent(in) :: il, iu, jl, ju
    character(len=*), intent(in) :: name
    integer(kind=int32), dimension(:, :), pointer :: arrayPtr
    integer :: status
    
    allocate(arrayPtr(il : iu, jl : ju), stat = status)
    call CheckAllocation(status, name)
    arrayPtr = 0
end function Allocate2DInt32

!***************************
! Function AllocateSingle
!***************************
function AllocateSingle(il, iu, name) result(arrayPtr)
    integer, intent(in) :: il, iu
    character(len=*), intent(in) :: name
    real(kind=single), dimension(:), pointer :: arrayPtr
    integer :: status
    
    allocate(arrayPtr(il : iu), stat = status)
    call CheckAllocation(status, name)
    arrayPtr = 0.0
end function AllocateSingle

!***************************
! Function Allocate2DSingle
!***************************
function Allocate2DSingle(il, iu, jl, ju, name) result(arrayPtr)
    integer, intent(in) :: il, iu, jl, ju
    character(len=*), intent(in) :: name
    real(kind=single), dimension(:, :), pointer :: arrayPtr
    integer :: status
    
    allocate(arrayPtr(il : iu, jl : ju), stat = status)
    call CheckAllocation(status, name)
    arrayPtr = 0.0
end function Allocate2DSingle

!***************************
! Function AllocateDouble
!***************************
function AllocateDouble(il, iu, name) result(arrayPtr)
    integer, intent(in) :: il, iu
    character(len=*), intent(in) :: name
    real(kind=double), dimension(:), pointer :: arrayPtr
    integer :: status
    
    allocate(arrayPtr(il : iu), stat = status)
    call CheckAllocation(status, name)
    arrayPtr = 0.0
end function AllocateDouble

!***************************
! Function Allocate2DDouble
!***************************
function Allocate2DDouble(il, iu, jl, ju, name) result(arrayPtr)
    integer, intent(in) :: il, iu, jl, ju
    character(len=*), intent(in) :: name
    real(kind=double), dimension(:, :), pointer :: arrayPtr
    integer :: status
    
    allocate(arrayPtr(il : iu, jl : ju), stat = status)
    call CheckAllocation(status, name)
    arrayPtr = 0.0
end function Allocate2DDouble


!***************************
! Function AllocateChar
!***************************
function AllocateChar(length, size, name) result(arrayPtr)
    integer, intent(in) :: length, size
    character(len=*), intent(in) :: name
    character(len=length), pointer, dimension(:) :: arrayPtr
    integer :: status
    
    allocate(arrayPtr(1 : size), stat = status)
    call CheckAllocation(status, name)

end function AllocateChar





!***************************************************
! CheckAllocation
!***************************************************
subroutine CheckAllocation(status, string)
    integer (kind=int32), intent(in) ::  status
    character(len=*), intent(in) :: string
    character(len=100) :: comment
   
    if(status .ne. 0) then
       write(comment, '(a32, a20)') "Could not allocate memory for : ", string 
       write (errstr, *) comment
       call fatal("allocErr") 
    end if
    noAllocations = noAllocations + 1
end subroutine CheckAllocation
!***************************************************
! CheckDeallocation
!***************************************************
subroutine CheckDeallocation(status, string)
    integer (kind=int32), intent(in) ::  status
    character(len=*), intent(in) :: string
    character(len=100) :: comment
   
    if(status .ne. 0) then
       write(comment, '(a31, a20)') "Failed to release memory for : ", trim(string) 
       write (errstr, *) comment
       call warning("allocErr", comment) 
    end if
    noDeallocations = noDeallocations + 1
end subroutine CheckDeallocation 

!***************************************************************
! Function MedianValue
! Find the median of X(1), ... , X(N), using as much of the quicksort
! algorithm as is needed to isolate it.
! N.B. On exit, the array X is partially ordered.
!**********************************************************************
real(kind=single) Function MedianValue(data, n)

    IMPLICIT NONE
    INTEGER, INTENT(IN)                :: n
    REAL(kind=single), INTENT(IN), DIMENSION(1 : n) :: data
    
    !************************************
    ! Local variables
    !***********************************

    REAL(kind=single)    :: temp, xhi, xlo, xmax, xmin
    LOGICAL (kind=bool) :: odd
    INTEGER (kind=int32) :: hi, lo, nby2, nby2p1, mid, i, j, k
    real(kind=single), dimension(:), allocatable :: x

  
    nby2 = n / 2
    nby2p1 = nby2 + 1
    odd  = .true.
    !******************************************************
    ! HI & LO are position limits encompassing the median.
    !******************************************************
    IF (n == 2 * nby2) odd = .false.
    lo = 1
    hi = n
    IF (n < 3) THEN
        IF (n < 1) THEN
            MedianValue = 0.0_double
            RETURN
        END IF
        MedianValue = data(1)
        IF (n == 1) RETURN
        MedianValue = 0.5 * (MedianValue + data(2))
        RETURN
    END IF


    !**********************************************
    ! Copy the data array to another one so that 
    ! original one is unchanged
    !**********************************************

    allocate(x(1 : n))
    x = data

    !********************************************
    ! Find median of 1st, middle & last values
    !********************************************
10 mid = (lo + hi)/2
    MedianValue = x(mid)
    xlo = x(lo)
    xhi = x(hi)
    IF (xhi < xlo) THEN          ! Swap xhi & xlo
        temp = xhi
        xhi = xlo
        xlo = temp
    END IF
    IF (MedianValue > xhi) THEN
        MedianValue = xhi
    ELSE IF (MedianValue < xlo) THEN
        MedianValue = xlo
    END IF
    !****************************************************
    ! The basic quicksort algorithm to move all values 
    ! <= the sort key (XMED) to the left-hand end, and 
    ! all higher values to the other end.
    !****************************************************
    i = lo
    j = hi
50  DO
        IF (x(i) >= MedianValue) EXIT
        i = i + 1
    END DO
    DO
        IF (x(j) <= MedianValue) EXIT
        j = j - 1
    END DO
    IF (i < j) THEN
        temp = x(i)
        x(i) = x(j)
        x(j) = temp
        i = i + 1
        j = j - 1

        !***************************************
        ! Decide which half the median is in.
        !***************************************

        IF (i <= j) GO TO 50
    END IF

    IF (.NOT. odd) THEN
        IF (j .eq. nby2 .AND. i .eq. nby2p1) GO TO 130
        IF (j < nby2) lo = i
        IF (i > nby2p1) hi = j
        IF (i /= j) GO TO 100
        IF (i == nby2) lo = nby2
        IF (j == nby2p1) hi = nby2p1
    ELSE
        IF (j < nby2p1) lo = i
        IF (i > nby2p1) hi = j
        IF (i /= j) GO TO 100
        !****************************************
        ! Test whether median has been isolated.
        !****************************************
        IF (i .eq. nby2p1) then
            deallocate(x)
            RETURN
        end if
    END IF
100 IF (lo .lt. hi - 1) GO TO 10

    IF (.NOT. odd) THEN
        MedianValue = 0.5 * (x(nby2) + x(nby2p1))
        deallocate(x)
        RETURN
    END IF
    temp = x(lo)
    IF (temp .gt. x(hi)) THEN
        x(lo) = x(hi)
        x(hi) = temp
    END IF
    MedianValue = x(nby2p1)
    deallocate(x)
    RETURN

    !***************************************************************
    ! Special case, N even, J = N/2 & I = J + 1, so the median is
    ! between the two halves of the series.   Find max. of the first
    ! half & min. of the second half, then average.
    !***************************************************************
    130 xmax = x(1)
    DO k = lo, j
        xmax = max(xmax, x(k))
    END DO
    xmin = x(n)

    DO k = i, hi
        xmin = min(xmin, x(k))
    END DO

    MedianValue = 0.5 * (xmin + xmax)
    deallocate(x)
END function MedianValue


end module omdetectnumerics_mod









