!**************************************************************
! module omdetectPhotometry_mod (10th July 2001)
!
! Module containing routines, etc,
! for doing aperture photometry on each source. I has been added
! because the counts that initially computed for each source (at the same
! time that the centroided positions for each source are obtained) is not
! consistent with the CAL calibration of source counts to magnitudes.
! It also sets the source confusion flag and the source
! quality flag. 
!
!**************************************************************
module omdetectPhotometry_mod
    
    use param
    use cal
    use dal
    
    use omdetectconstants_mod
    use omdetectccddata_mod
    use omdetectnumerics_mod 
 
  contains
!****************************************************************************************
! subroutine AperturePhotometry
! Performs aperture photometry on a set of sources
! using a circular aperture of a given pixel radius
! Parameters-
!     nx and ny - the dimensions of the images
!     image - the image
!     sourceQualityImage - describes the quality of each image pixel
!     photometryRadius - The aperture radius for doing the photometry
!     noSources - number of sources
!     xSource - array of source x-positions
!     ySource - array of source y-positions
!     psfXwidth - array holding the computed x extents of the sources (pixels).
!     psfYwidth - array holding the computed y extents of the sources (pixels).
!     fSource - array to hold the counts for each source.
!               Contains the Jon Mittaz counts on entry
!               and where possible these will be replaced 
!               by the aperture photometry values.
!     fBackGround - array to hold the background 
!               counts for each source. 
!               Contains the Jon Mittaz counts on entry
!               and where possible these will be replaced 
!               by the aperture photometry values.
!     sourceExtendedFlag - array containg either 0 (not extended source)
!                or 1 (extended source)
!     sourceConfusionFlag - array containing either 1 - no source within
!                           2, source within aperture but outside minimum aperture
!                           radius, or 3 - source within minimum apertuire radius.
!     sourceQualityFlag - Set to 1 if source region has a bad-quality pixel- 0 otherwise
!     sourceApertureRadius - the aperture radius that was used for     
!                      each source 
!****************************************************************************************
subroutine AperturePhotometry(   &
    nx,                      &  ! The x-extent of the images 
    ny,                      &  ! The y-extent of the images 
    image,                   &  ! The image (including background)
    sourceQualityImage,      &  ! The pixel-quality image array 
    photometryRadius,        &  ! The aperture radius (pixels) to be used for doing photometry
    noSources,               &  ! The number of sources
    xSource,                 &  ! The x-pixel position of each source 
    ySource,                 &  ! The y-pixel position of each source 
    psfXwidth,               &  ! The full-width at half-maximum (pixels) of each source along the major axis 
    psfYwidth,               &  ! The full-width at half-maximum (pixels) of each source along the minor axis
    fSource,                 &  ! Counts for each source (modified)
    fSourceError,            &  ! Error in counts (modified)
    sourceSignificance,      &  ! The significance of each source (modified)
    fBackGround,             &  ! The background counts of each source (modified)
    sourceExtendedFlag,      &  ! Flag for each source to say if source is extended, or not
    sourceConfusionFlag,     &  ! Flag for each source to say if source within aperture (modified)
    sourceQualityFlag,       &  ! Flag for each source to say if a pixel within the aperture was bad (modified)
    sourceApertureRadius)       ! The aperture radius (pixels) for each source that was used for doing photometry
 
    !***************************************
    ! Input and unmodified variables
    !***************************************
    integer (kind=int32), intent(in) :: nx, ny                      
    real(kind=single), dimension(0 : nx - 1, 0 : ny - 1)               :: image                        
    logical(kind=bool), intent(in), dimension(0 : nx - 1, 0 : ny - 1)  :: sourceQualityImage 
    real(kind=single), intent(in)                                      :: photometryRadius   
    integer (kind=int32), intent(in)                                                :: noSources 
    real(kind=single), intent(in), dimension(1 : noSources)            :: xSource, ySource       
    integer(kind=int8), intent(in), dimension(1 : noSources)           :: sourceExtendedFlag   
  
    !***************************************
    ! Input and modified variables
    !***************************************
    real(kind=single), intent(inout), dimension(1 : noSources)  :: psfXwidth, psfYwidth 
    real(kind=single), intent(inout), dimension(1 : noSources)  :: fSource            
    real(kind=single), intent(inout), dimension(1 : noSources)  :: fSourceError        
    real(kind=single), intent(inout), dimension(1 : noSources)  :: sourceSignificance  
    real(kind=single), intent(inout), dimension(1 : noSources)  :: fBackGround         

    !***************************************
    ! Output variables
    !***************************************
    real(kind=single), intent(out), dimension(1 : noSources)    :: sourceApertureRadius
    integer(kind=int8), intent(out), dimension(1 : noSources)   :: sourceConfusionFlag  
    integer(kind=int8), intent(out), dimension(1 : noSources)   :: sourceQualityFlag    
 
    !**************************************
    ! Local variables
    !**************************************
    integer (kind=int32)           :: i, k, status, type
    real(kind=double) :: sourceCounts      ! The aperture counts for a source (including background)
    real(kind=double) :: backGroundCounts  ! The background counts for a source    
    real(kind=single) :: maxRadius         

    !***********************************************************
    ! NB Set minSignificance to zero if no counts extrapolation
    ! to be used ! NB until the CAL calibration data is in place
    ! the full aperture will be used on all sources, unless
    ! another source lies within the aperture. If the source 
    ! significance is less than this value then
    ! the photometric counts will be extrapolated using
    ! the point-spread-function
    !***********************************************************
    real(kind=single), parameter :: minSignificance = 0.0  
    integer (kind=int32) :: index

    if (noSources .lt. 1) return ! Return if no sources

    !***************************************************
    ! Allocate memory for arrays and check if successful
    !***************************************************

    allocate(CCDImage(0 : nx - 1, 0 : ny - 1), stat = status)
    call CheckAllocation(status, "CCDIMage (Subroutine AperturePhotometry)")

  
    allocate(qualityImage(0 : nx - 1, 0 : ny - 1), stat = status)
    call CheckAllocation(status, "qualityImage (Subroutine AperturePhotometry)")
    qualityImage = sourceQualityImage 

    allocate(qualityFlag(1 : noSources), stat = status)
    call CheckAllocation(status, "qualityFlag (Subroutine AperturePhotometry)")
    qualityFlag = 0 

    allocate(sortIndex(1 : noSources), stat=status)
    call CheckAllocation(status, "sortIndex (Subroutine AperturePhotometry)")

    allocate(sourceCoords(1 : noSources), stat=status)
    call CheckAllocation(status, "sourceCoords (Subroutine AperturePhotometry)")


    allocate(objectWidth(1 : noSources), stat = status)
    call CheckAllocation(status, "objectWidth (Subroutine AperturePhotometry)")
 

    !**************************************************************
    ! Set the minimum aperture radius to half the required aperture
    ! for now- until CAL calibratiomn data in place
    !**************************************************************
    minimumApertureRadius = photometryRadius * 0.5
  

    !************************************************
    ! Compute the delta x and delta y offsets that
    ! are used to compute the background from an
    ! annulus region around the source
    !************************************************
    call SetUpBackgroundOffsets(photometryRadius) 
  
    
    !**************************************************
    ! Obtain an index of the source positions in
    ! increasing x-pixel position. NB xSource is not
    ! sorted. sortIndex(i) will give the index of the 
    ! array element that gives the positions in increasing
    ! x-position. This is used for obtaining the distance
    ! from one source to its nearest neighbour. 
    !**************************************************
    call IndexSort(xSource, sortIndex, noSources)  

    !*******************************************************
    ! sourceCoords will contain the position of the sources
    ! in increasing x-position- this array is used to obtain
    ! the maximum-sized aperture for doing photometry.
    !*******************************************************
 
    forall (i = 1 : noSources)
        sourceCoords(i)%x = xSource(sortIndex(i))
        sourceCoords(i)%y = ySource(sortIndex(i))

        !**************************************************************************
        ! objectWidth will store the maximum of the x and y extents of the sources
        ! * 3, or if this is less than the aperture radius, to the aperture radius.
        !**************************************************************************      
        objectWidth(i) = 3.0 * max(psfXwidth(sortIndex(i)), psfYwidth(sortIndex(i)))
        objectWidth(i) = max(objectWidth(i), photometryRadius)          
   
    end forall

  
    !********************************************************************************
    ! Loop through all the sources doing aperture photometry on each one (if possible)
    !********************************************************************************

    xLeft = 0.0
    yBottom = 0.0
    xRight = real(nx)
    yTop = real(ny)
 
  
    k = 0
    do i = 1,  noSources
      
        index = sortIndex(i)
   
 
        !*********************************************
        ! For the moment do photometry on
        ! extended sources
        !*********************************************  

        if (sourceExtendedFlag(index) .lt. 10) then 
            sourceNumber = index

    
            !********************************************
            ! Find the aperture radius that can be used
            ! for the photometry. If type is 1 then
            ! can use the required aperture, 2 - have
            ! to use a smaller one, 3 can't do photometry
            !********************************************

            apertureRadius = photometryRadius
            maxRadius = GetMaximumAperture(noSources, i, type)          

            select case (type)     !************************************************
                case (1, 2)        ! 1 - No other objects within aperture, 
                                   ! 2 - Another object within aperture, but outside 
                                   ! minimum aperture radius- use aperture that 
                                   ! excludes this source
                                   !************************************************

                    if (sourceSignificance(index) .lt. minSignificance) then ! Use a smaller aperture
                        maxRadius = minimumApertureRadius
                    end if
    
                    sourceCounts = GetApertureCounts(image, nx, ny, xSource(index), ySource(index), &
                                        maxRadius) 
                    backGroundCounts = GetBackGroundCounts(xSource(index), ySource(index), maxRadius)
                    sourceCounts = sourceCounts - backGroundCounts
            
                    sourceApertureRadius(index) = maxRadius * real(binX)
             

                    !*****************************************
                    ! Recalculate the source significance
                    !*****************************************   
                    sourceSignificance(index) = ComputeSourceSignificance(sourceCounts, backGroundCounts)
           

                    !*******************************************************
                    ! Replace the old source counts with the new value
                    !*******************************************************
                    fSource(index) = sourceCounts            ! Source counts
                    fSourceError(index) = ComputeSourceCountsError(sourceCounts, backGroundCounts)
                    fBackGround(index) = backGroundCounts    ! Background counts

  
                case (3)   ! Object within aperture radius / 2.0  ! Leave as original for now 

                    sourceApertureRadius(index) = photometryRadius * real(binX)
                    
       
            end select
        else

             sourceApertureRadius(index) = photometryRadius * real(binX)                    
             type = 1
        end if        

  
        !****************************************************
        ! Set the source confusion flag - 0 isolated source
        ! 1 Source within aperture radius and greater than minimum aperture radius
        ! 2 Source within minimum aperture radius
        !****************************************************         
        sourceConfusionFlag(index) = type
    end do

 
    deallocate(backgroundOffset, stat = status)
    call CheckDeallocation(status, "backgroundOffset (Subroutine AperturePhotometry)")

    deallocate(backGroundPixel, stat = status)
    call CheckDeallocation(status, "backGroundPixel (Subroutine AperturePhotometry)")

    deallocate(CCDImage, stat = status)
    call CheckDeallocation(status, "CCDIMage (Subroutine AperturePhotometry)")

    deallocate(qualityImage, stat = status)
    call CheckDeallocation(status, "qualityImage (Subroutine AperturePhotometry)")
 
    sourceQualityFlag = qualityFlag
    deallocate(qualityFlag, stat = status)
    call CheckDeallocation(status, "qualityFlag (Subroutine AperturePhotometry)")
 
    deallocate(sortIndex, stat = status)
    call CheckDeallocation(status, "sortIndex (Subroutine AperturePhotometry)")

    deallocate(sourceCoords, stat = status)
    call CheckDeallocation(status, "sourceCoords (Subroutine AperturePhotometry)")

    deAllocate(objectWidth, stat = status)
    call CheckDeallocation(status, "objectWidth (Subroutine AperturePhotometry)")

end subroutine AperturePhotometry

!*********************************************
! Function GetApertureCounts
! Computes the counts within a given aperture
! Returns the value via the function name
!*********************************************
real(kind=double) function GetApertureCounts(&
                 image,  &  ! The 2-d image to be used to get the counts within the aperture
                    nx,  &  ! The x-extent of the image
                    ny,  &  ! The y-extent of the image
                    x0,  &  ! The x-pixel coordinate of the centre of the aperture
                    y0,  &  ! The y-pixel coordinate of the centre of the aperture
                radius)     ! The aperture radius (pixels) for summing the counts  
    implicit none

    !*********************************************
    ! Input (unmodfied) variables
    !*********************************************
    integer (kind=int32), intent(in) :: nx
    integer (kind=int32), intent(in) :: ny
    real(kind=single), intent(in), dimension(0 : nx - 1, 0 : ny - 1) :: image
    real(kind=single), intent(in) :: x0
    real(kind=single), intent(in) :: y0
    real(kind=single), intent(in) :: radius
   
    !********************************************
    ! Local variables
    !*******************************************
    integer (kind=int32) :: i, j, i0, j0, r
    integer (kind=int32) :: i1, i2, j1, j2



    CCDImage = image

    apertureRadius = radius        
    il = 0
    iu = nx - 1
    jl = 0
    ju = ny - 1
        
    i0 = int(x0)
    j0 = int(y0)
    xLeft = 0.0
    yBottom = 0.0
    xRight = real(nx)
    yTop = real(ny)
 
   
    r = ceiling(radius)

    !************************************************
    ! Determine upper and lower array indices for
    ! summing counts
    !***********************************************

    i1 = i0 - r - 1
    il = max(0, il)
  
    i2 = i0 + r + 1
    i2 = min(iu, i2)  

    j1 = j0 - r - 1
    j1 = max(0, j1)

    j2 = j0 + r + 1
    j2 = min(ju, j2)
 
    GetApertureCounts = 0.0_double 

    do i = i1, i2
        do j = j1, j2          
            GetApertureCounts = GetApertureCounts + PixelValue(x0, y0, i, j)
        end do
    end do

end function GetApertureCounts       
!********************************************
! Function PixelValue
! Purpose - returns the value of the pixel
! at the given position weighted by the area
! of the aperture contained within the pixel
!*********************************************
real(kind=double) function PixelValue(x0, y0, iIndex, jIndex)
    implicit none   
    real (kind=single), intent(in) :: x0, y0
    real (kind=single):: r1, r2, r3, r4
    real (kind=single) :: x, y, x1, x2, y1, y2
    real (kind=single) :: minRadius, maxRadius
    
    integer (kind=int32), intent(in) :: iIndex, jIndex
 
    x = real(iIndex)
    y = real(jIndex)
 
    x1 = x - x0
    y1 = y - y0
    x2 = x1 + 1.0
    y2 = y1 + 1.0

    r1 = x1 * x1 + y1 * y1
    minRadius = r1
    maxRadius = r1
 
    r2 = x2 * x2 + y1 * y1
    minRadius = min(minRadius, r2)
    maxRadius = max(maxRadius, r2)

    r3 = x1 * x1 + y2 * y2
    minRadius = min(minRadius, r3)
    maxRadius = max(maxRadius, r3)

    r4 = x2 * x2 + y2 * y2
    minRadius = sqrt(min(minRadius, r4))
    maxRadius = sqrt(max(maxRadius, r4))
   
    if (apertureRadius .le. minRadius) then
        PixelValue =  0.0_double
        return 
    end if

    !*****************************************
    ! Check that the pixel is okay to use
    !*****************************************

    if (qualityImage(iIndex, jIndex)) then ! Bad pixel- don't use- set source quality flag
        qualityFlag(sourceNumber) = 1
        PixelValue = 0.0_double
        return
    end if
 
    if (apertureRadius  .ge. maxRadius) then
              
        PixelValue = CCDImage(iIndex, jIndex)
    else
  
        pixelValue = EffectivePixelArea(x1, y1) * CCDImage(iIndex, jIndex)
              
    end if

end function PixelValue
!**************************************************
! Function EffectivePixelArea
! Computes the area of a pixel contained
! within the aperture and returns the value
! via the name
! This will shortly be replaced by a better routine
!***************************************************
real(kind=double) function EffectivePixelArea(x1, y1)
    implicit none
    real(kind=single), intent(in) :: x1, y1
    real(kind=single) :: radiusSqr
    real(kind=single) :: xPos, yPos
    
    integer (kind=int32), parameter :: maxTimes = 99 * 99
    integer (kind=int32) :: nIn
    integer (kind=int32) :: i, j
    real (kind=single) :: rsq

    rsq = apertureRadius * apertureRadius
    nIn = 0
  
    nIn = 0

    do i = 2, 100

        xPos = x1 + 0.01 * real(i - 1)
        do j = 2, 100  
            yPos = y1 + 0.01 * real(j - 1)
        
            radiusSqr = xPos * xPos + yPos * yPos 
            if (radiusSqr .le. rsq) then                         
                nIn = nIn + 1
            end if
        end do
    end do

    EffectivePixelArea = dble(nin) / dble(maxTimes) 

end function EffectivePixelArea
!***********************************************
! ComputeSourceCountsError
! Computes the source counts error and returns 
! the value via the name
!***********************************************
real(kind=double) function ComputeSourceCountsError(sourceCounts, backGroundCounts)
    real(kind=double), intent(in) :: sourceCounts       ! The background-subtracted source counts
    real(kind=double), intent(in) :: backGroundCounts   ! The background counts
    
    ComputeSourceCountsError = sqrt(sourceCounts + 2.0_double * backGroundCounts)
 
end function ComputeSourceCountsError
!***********************************************
! ComputeSourceSignificance
! Computes the source significance and returns
! the value via the name
!***********************************************
real(kind=double) function ComputeSourceSignificance(sourceCounts, backGroundCounts)
    real(kind=double), intent(in) :: sourceCounts     ! The background-subtracted source counts
    real(kind=double), intent(in) :: backGroundCounts ! The background counts

    ComputeSourceSignificance = sourceCounts  / sqrt(backGroundCounts)
end function ComputeSourceSignificance


!********************************************************
! Subroutine CreatePSFImage
! Purpose - Creates point-spread function image centred
! at x0, y0
!********************************************************
subroutine CreatePSFImage(x0, y0, countFrameRatio)
    real(kind=single), intent(in) :: x0, y0 
    real(kind=single), intent(in) :: countFrameRatio
    real(kind=single), dimension(:, :), pointer :: psf ! To hold the point-spread function
    integer (kind=int32) ::  xRes, yRes
    integer (kind=int32) :: status      
    
    !********************************
    ! Get the point-spread function
    !********************************
    psfNx = 10
    psfNy = 10
   
    xres = 5
    yres = 5
    
    call CAL_omGetPSFmap(dble(x0), dble(y0), psfNx, psfNx, psfNy, psfNy, &
                         xRes, yRes, dble(countFrameRatio), psf)
 
    psfNx = ubound(psf, 2) + 1
    psfNy = ubound(psf, 1) + 1
   
    if (.not. allocated( psfImage)) then
        allocate(psfImage(0 : psfNx - 1, 0 : psfNy - 1), stat=status)
        call CheckAllocation(status, "psfImage (Subroutine AperturePhotometry)")
    end if

    psfImage = psf
   
    call CAL_releaseMemory(psf)

end subroutine CreatePSFImage
!************************************************************
! Subroutine HeapSort
! Sorts  one-dimensional array(1.. size) of type POINTS, 
! into ascending radius order.
!************************************************************
subroutine HeapSort1(array, size)
    integer(kind = int32), intent(in) :: size ! The size of the array
    type(POINT), intent(inout), dimension( : ) :: array  ! The array to be sorted
    integer(kind = int32) :: i, ir, j, l
    type(POINT) :: rra
  
    if (size .lt. 2) return

    l = size / 2 + 1
    ir = size   

    do 
        if (l .gt. 1) then
            l = l - 1
            rra = array(l)    
        else
            rra = array(ir)
            array(ir) = array(1)
            ir = ir - 1
            if (ir .le. 1) then
                array(1) = rra
                exit
            end if
        end if

        i = l
        j = l + l
        do while(j .le. ir)
       
            if (j .lt. ir) then
                if(array(j)%radius .lt. array(j + 1)%radius) j = j + 1
            end if
            if (rra%radius .lt. array(j)%radius) then
                array(i) = array(j)
                i = j
                j = j + j
            else
                exit
            end if
            
        end do   
        array(i) = rra    
    end do
end subroutine HeapSort1

!************************************************************
! Subroutine IndexSort
! Sorts  one-dimensional array(1.. size) int ascending order.
! An index of the sorted order relative to the original
! order is returned.
!************************************************************ 
subroutine IndexSort(array, index, size)
    integer (kind=int32), intent(in) :: size
    real(kind=single), intent(in), dimension(1 : size) :: array
    integer (kind=int32), intent(out), dimension(1 : size) :: index

    integer (kind=int32) :: i, ir, j, l, irra
    real(kind=single) :: rra
    if (size .eq. 0) return
    if (size < 2) then
        index(1) = 1
    end if

    l = size / 2 + 1
    ir = size   
    do i = 1, size
        index(i) = i
       
    end do   

    do 
        if (l > 1) then
            l = l - 1
            rra = array(index(l))   
            irra = index(l)    
        else
            rra = array(index(ir))        
            irra = index(ir)
          
            index(ir) = index(1)
            ir = ir - 1
            
            if (ir <= 1) then
             
                index(1) = irra
                exit
            end if
        end if

        i = l
        j = l + l
        do while(j <= ir)
       
            if (j < ir) then
                if(array(index(j)) < array(index(j + 1))) j = j + 1
            end if
            if (rra < array(index(j))) then
               
                index(i) = index(j)
                i = j
                j = j + j
            else
                exit
            end if
            
        end do   
          
        index(i) = irra
    end do
   
end subroutine IndexSort
!******************************************************************
! Function GetMaximumAperture
! Finds the maximum aperture size that can be used for photomery
! Returns the value via the function name
! Also return "type" -
! type - 1 = no object within aperture radius pixels
!        2 = object within aperture radius, but outside minimum aperture radius
!        3 = object within minimum aperture radius
!*******************************************************************
real(kind=single) function GetMaximumAperture(noSources, index, type)
    integer (kind=int32), intent(in) :: noSources        ! The number of sources    
    integer (kind=int32), intent(in) :: index            ! Index of source 
    integer (kind=int32), intent(out) :: type            ! Determines what kind of photometry can be performed
    TYPE(point) :: centre                   ! Coords of source
    real(kind=single) :: minimumSeparation  ! Minimum distance between source of interest and other sources
    real(kind=single) :: borderDistance     ! Distance to the edge of the image from the source 
    real(kind=single) :: okayDistance       ! Minimum allowed separation between source and nearest neighbour   
    real(kind=single) :: distance           ! Distance bewteen source and neighbour
    real (kind=single):: d, d1
    integer (kind=int32) :: k    
   
    centre = sourceCoords(index)

    okayDistance = apertureRadius  
    borderDistance = 1e8
    minimumSeparation = 1e6



    !*******************************************
    ! Check to see if object is near the
    ! edge of the image
    !*******************************************
    if (centre%x - xLeft .lt. apertureRadius) then
        borderDistance = centre%x - xLeft
       
    else if  (xRight - centre%x .lt. apertureRadius) then
       
        borderDistance = xRight - centre%x
    end if

    if (centre%y - yBottom .lt. apertureRadius) then
        borderDistance = min(centre%y - yBottom, borderDistance)
       
    else  if (yTop - centre%y .lt. apertureRadius) then
        borderDistance = min(yTop - centre%y, borderDistance)
        
    end if
        

    !*************************************************
    ! Find minimum distance between source of interest
    ! and sources to its left. NB sourceCoords contains
    ! the x-coordinates of the sources in increasing 
    ! order.
    !*************************************************
    if (index .gt. 1) then
        k = index - 1
    
        do while (k .ge. 1)
            d = centre%x - sourceCoords(k)%x - objectWidth(k)
            d = max(0.0, d)
            if (d .gt. okayDistance) exit
            if (sourceCoords(k)%y .ge. centre%y) then
                d1 = sourceCoords(k)%y - centre%y - objectWidth(k)
                d1 = max(0.0, d1)
                
            else
                d1 = centre%y - sourceCoords(k)%y - objectWidth(k)
                d1 = max(0.0, d1)
            end if


            distance = d * d + d1 * d1
            
            if (distance .lt. minimumSeparation) then
                minimumSeparation = distance
            end if
  

            k = k - 1
        end do
    end if

    !*************************************************
    ! Find minimum distance between source of interest
    ! and sources to its right.
    !*************************************************
    if (index .lt. noSources) then
        k = index + 1
        do while (k .le. noSources)
            d = sourceCoords(k)%x - objectWidth(k) - centre%x
            d = max(0.0, d)

            if (d .gt. okayDistance) exit

            if (sourceCoords(k)%y .ge. centre%y) then
                d1 = sourceCoords(k)%y - centre%y - objectWidth(k)
                d1 = max(0.0, d1)
            else
                d1 = centre%y - sourceCoords(k)%y - objectWidth(k)
                d1 = max(0.0, d1)
            end if

            distance = d * d + d1 * d1
            if (distance .lt. minimumSeparation) then
                minimumSeparation = distance
            end if
            k = k + 1
        end do
    end if

    minimumSeparation = sqrt(minimumSeparation)


    if (minimumSeparation .ge. okayDistance) then
 
        if (borderDistance .ge. apertureRadius) then
            type = 1
            GetMaximumAperture = apertureRadius

        else if (borderDistance .gt. minimumApertureRadius) then
            type = 2 
            GetMaximumAperture = borderDistance
        
        else
            type = 3    ! Can't do photometry on
            GetMaximumAperture = apertureRadius
 
        end if
        return
    end if
       
    minimumSeparation = min(minimumSeparation, borderDistance)

    if (minimumSeparation .ge. okayDistance) then
        type = 1
        GetMaximumAperture = apertureRadius

    else if (minimumSeparation .ge. minimumApertureRadius) then  
        type = 2 
        GetMaximumAperture = minimumSeparation   ! - apertureRadius
        
    else
        type = 3    ! Can't do photometry on
        GetMaximumAperture = 0.0
    end if
 
end function GetMaximumAperture

!*************************************
! Subroutine SetUpBackgroundOffsets
! Sets up an array of x and y pixel offsets
! from the centre of the aperture coordinates 
! to be used for computing the background 
! for each source
!***************************Rradius**********
subroutine SetUpBackgroundOffsets(radius)
    real(kind=single), intent(in) :: radius  ! Theb required photometry aperture radius (pixels)
    integer (kind=int32) :: i, j, il, iu, jl, ju
    integer (kind=int32) :: status
    integer (kind=int32) :: noBackGroundPixelsNeeded
    real(kind=single) :: r 
    real(kind=single) :: innerRadius  ! The inner radius of the annulus
    real(kind=single) :: outerRadius  ! The outer radius of the annulus



    !***************************************************************************
    ! For a good estimate of the background need to sample 3 times the number of 
    ! source pixels within the aperture radius.
    !***************************************************************************
    innerRadius = radius + 2.0  
    outerRadius = ceiling (sqrt(3.0  * radius * radius + innerRadius * innerRadius))  

    noBackGroundPixelsNeeded = ceiling(PI * (outerRadius * outerRadius - innerRadius * innerRadius))

    
    allocate(backgroundOffset(1 : noBackgroundPixelsNeeded + 10), stat=status)
    call CheckAllocation(status, "backGroundOffset (Subroutine AperturePhotometry)")

    allocate(backGroundPixel(1 : noBackgroundPixelsNeeded + 10), stat=status)
    call CheckAllocation(status, "backGroundPixel (Subroutine AperturePhotometry)")

    iu = ceiling(outerRadius)
    ju = iu
    il = -iu
    jl = il

    nBackgroundOffsets = 0
    do i = il, iu
        do j = jl, ju
            r  = sqrt(real(i * i + j * j))
            if (r .gt. innerRadius .and. r .le. outerRadius) then
                nBackgroundOffsets = nBackgroundOffsets + 1
                backGroundOffset(nBackgroundOffsets) = POINT(i, j, r)
            end if 
        end do
    end do
 
    !*******************************************************
    ! Sort these offsets into the order of increasing radius
    !******************************************************* 
    call HeapSort1(backGroundOffset, nBackgroundOffsets)
 
end subroutine SetUpBackgroundOffsets  
!*****************************************
! subroutine ComputePSFWidths
! Computes the semi-major and semi-minor
! axes of the point-spread function by
! fitting a 2-dimensional function to an
! image of the point-spread function
!*****************************************
subroutine ComputePSFWidths( &
    psfSemiA,  &    !  The computed semi-major axis of the psf 
    psfSemiB)       !  The computed semi-minor axis mof the psf
    real(kind=single), intent(out) :: psfSemiA, psfSemiB
    !******************************
    ! Local variables
    !******************************
    integer (kind=int32) :: i, j
    real (kind=single) :: parameter(6)  ! Stores the fitted parameters
    integer (kind=int32), pointer :: psfLevelImage(:, :)  ! Pointer to the psf image
    real(kind=single), dimension(:, :), allocatable ::  image  ! 2-d array to store the psf image

    integer (kind=int32) :: psfNx = 41
    integer (kind=int32) :: psfNy = 41     



    !******************************************************
    ! Create the PSF  -image map
    !*****************************************************
    call CreatePSFImage(1024.5, 1024.5, 0.0)
   
    allocate(psfLevelImage(1 : psfNx, 1 : psfNy))
    psfLevelImage = 1


    allocate(image(1 : psfNx, 1 : psfNy))

    forall(i = 1 : psfNx)
        forall(j = 1 : psfNy)  image(i, j) = psfImage(i - 1, j - 1)
    end forall 

 
    !************************************************************
    ! Fit a 2-dimensional gaussian to the point-spread function
    !************************************************************
    call fitGaussian(image, psfNx, psfNy, psfLevelImage, &
                      1, psfNx, 1, psfNy, parameter)

    !*****************************************************
    ! The x and y semi-major axes in pixels
    !*****************************************************
    psfSemiA = parameter(2) / 5.0
    psfSemiB = parameter(3) / 5.0 
    
    deallocate(image)
    deallocate(psfImage)
    deallocate(psfLevelImage)

end subroutine ComputePSFWidths
!*********************************************
! Function GetBackGroundCounts
! Computes the background counts within the given
! annulus, using the median counts/pixels obtained
! from an annulus region. 
! Return the value via the function name 
!*********************************************
real (kind=double) function GetBackGroundCounts( &
    x0,     &   ! The centre x-pixel coordinate of the aperture
    y0,     &   ! The centre y-pixel corrdinate of the aperture
    radius)     ! The aperture radius
    implicit none
    !***************************************
    ! Input variables
    !***************************************
    real(kind=single), intent(in) :: radius
    real(kind=single), intent(in) :: x0
    real(kind=single), intent(in) :: y0
    !***************************************
    ! Local variables
    !***************************************
    integer (kind=int32) :: i, j, k, i0, j0
    integer (kind=int32) :: n 
    integer (kind=int32) :: ndata  ! The number of background pixels used
    real(kind=double) :: median 
    real(kind=double) :: cutOff  
!    real(kind=single) :: mode 

    !*****************************************************
    ! Scan the annulus region and add pixel values to the 
    ! backGroundPixel array
    !*****************************************************

    backGroundPixel = 0.0
    i0 = int(x0)
    j0 = int(y0)
    nData = 0    

    do k = 1, nBackgroundOffsets 
        i = i0 + backGroundOffset(k)%x
        j = j0 + backGroundOffset(k)%y

        if (i .ge. il .and. i .le. iu .and. j .ge. jl .and. j .le. ju) then
            if (CCdImage(i, j) .ne. pixelIgnoreFlag) then
                nData = nData + 1
                backGroundPixel(nData) = CCDImage(i, j)   
            end if
        end if
  
    end do
  

    !************************************************
    ! Compute the median background pixel value
    !************************************************
    if (nData .gt. 0) then

        median = MedianValue(backGroundPixel, nData)
!         median = TheMode(backGroundPixel, nData, 2.0)
        GetBackgroundCounts = median * PI * radius * radius
 
    else
        GetBackgroundCounts = 0.0_double
        return
    end if


 
    !**************************************************
    ! Reject pixel values > 1.3 * median value
    !**************************************************
    cutOff = median * 1.3
   
    n = 0
    do i = 1, nData
 
        if (backGroundPixel(i) .le. cutOff) then
            n = n + 1
            backGroundPixel(n) = backGroundPixel(i)
        end if
    end do 

    !**************************************************
    ! Recompute the median and calculate the counts
    ! contained within the aperture
    !*************************************************

    if (n .gt. 0 .and. n .lt. nData) then 
        median = MedianValue(backGroundPixel, n)  
        GetBackgroundCounts = median * PI * radius * radius
    end if

end function GetBackgroundCounts       

!******************************************
! PSFFractionOfTotalCounts
! Returns the fraction of the total counts
! contained within the given radius
!******************************************
real(kind=double) function & 
    PSFFractionOfTotalCounts(x0, y0, radius, countFrameRatio)
    real(kind=double), intent(in) :: x0
    real(kind=double), intent(in) :: y0
    real(kind=double), intent(in) :: radius
    real(kind=double), intent(in) :: countFrameRatio
    type(OmPsfT) :: psf
 
    psf = CAL_omGetPSF(x0, y0, countFrameRatio)

    PSFFractionOfTotalCounts = CAL_omPsfEncircledEnergy(psf, radius)
!write(*, *) "Function FractionOfTotalCounts- radius=", radius, " fraction=", CAL_omPsfEncircledEnergy(psf, radius) 

end function PSFFractionOfTotalCounts 

!******************************************
! PSFRadiusContainingCounts
! Returns the radius containing
! the specified fraction of the total counts
!******************************************
real(kind=double) function & 
    PSFRadiusContainingCounts(x0, y0, fraction, countFrameRatio)
    real(kind=double), intent(in) :: x0
    real(kind=double), intent(in) :: y0
    real(kind=double), intent(in) :: fraction
    real(kind=double), intent(in) :: countFrameRatio
    type(OmPsfT) :: psf
 
    psf = CAL_omGetPSF(x0, y0, countFrameRatio)

    PSFRadiusContainingCounts = CAL_omPsfCircleRadius(psf, fraction)

end function PSFRadiusContainingCounts

end module omdetectPhotometry_mod



























